# -*- coding: UTF-8 -*-
# @Time    : 17-5-9
# @File    : challenge.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from antgo.resource.html import *
from antgo.ant.base import *
from antgo.dataflow.common import *
from antgo.measures.statistic import *
from antgo.task.task import *
from antgo.utils import logger
from antgo.dataflow.recorder import *
from antgo.measures.deep_analysis import *
from antgo.measures.significance import *
import shutil
import tarfile
from datetime import datetime
from antgo.measures import *
from antgo.measures.yesno_crowdsource import *
import traceback
import zlib
from antgo.dataflow.dataset.proxy_dataset import *


class AntChallenge(AntBase):
  def __init__(self, ant_context,
               ant_name,
               ant_data_folder,
               ant_dump_dir,
               ant_token,
               ant_task_config=None,
               ant_dataset=None,
               ant_task_benchmark=None,
               **kwargs):
    super(AntChallenge, self).__init__(ant_name, ant_context, ant_token, **kwargs)
    self.ant_data_source = ant_data_folder
    self.ant_dump_dir = ant_dump_dir
    self.ant_context.ant = self
    self.ant_task_config = ant_task_config
    self.ant_dataset = ant_dataset
    self.ant_task_benchmark = ant_task_benchmark
    self.context.devices = self.context.params.system['devices']
    self._running_dataset = None
    self._running_task = None
    self.task_running_statictic = {}

  @property
  def running_dataset(self):
    return self._running_dataset

  @property
  def running_task(self):
    return self._running_task

  def evaluate(self):
    if not self.context.recorder.is_measure:
      return

    logger.info('Start evaluation process.')
    evaluation_measure_result = []
    with safe_recorder_manager(RecordReader(self.context.recorder.dump_dir)) as record_reader:
      for measure in self.running_task.evaluation_measures:
        if measure.crowdsource:
          # start crowdsource server
          measure.dump_dir = os.path.join(infer_dump_dir, measure.name, 'static')
          if not os.path.exists(measure.dump_dir):
            os.makedirs(measure.dump_dir)

          measure.experiment_id = self.context.experiment_uuid
          measure.app_token = self.token
          logger.info('Launch crowdsource evaluation server.')
          crowdsrouce_evaluation_status = measure.crowdsource_server(record_reader)
          if not crowdsrouce_evaluation_status:
            logger.error('Couldnt finish crowdsource evaluation server.')
            continue

          # using crowdsource evaluation
          result = measure.eva()
          # TODO: support bootstrap confidence interval for crowdsource evaluation
        else:
          # evaluation
          record_generator = record_reader.iterate_read('predict', 'groundtruth')
          result = measure.eva(record_generator, None)

          # compute confidence interval
          if measure.is_support_rank and getattr(self.running_task, 'confidence_interval', False):
            confidence_interval = bootstrap_confidence_interval(record_reader, time.time(), measure, 10)
            result['statistic']['value'][0]['interval'] = confidence_interval
          elif measure.is_support_rank:
            result['statistic']['value'][0]['interval'] = (
            result['statistic']['value'][0]['value'], result['statistic']['value'][0]['value'])

        evaluation_measure_result.append(result)

      # #########################
      # roc_auc_measure = {'statistic': {'name': 'roc_auc',
      #                                  'value': [{'name': 'ROC', 'value': [[[0, 3], [1, 0], [2, 6]],
      #                                                                      [[0, 8], [2, 3], [3, 7]]],
      #                                             'type': 'CURVE', 'x': 'FP', 'y': 'TP',
      #                                             'legend': ['class-0', 'class-1']},
      #                                            {'name': 'AUC', 'value': [0.1, 0.2], 'type': 'SCALAR', 'x': 'class',
      #                                             'y': 'AUC'}]}}
      #
      # voc_measure = {'statistic': {'name': 'voc',
      #                              'value': [{'name': 'MAP', 'value': [18.0, 9.0, 20.0], 'type': 'SCALAR', 'x': 'class',
      #                                         'y': 'Mean Average Precision'},
      #                                        {'name': 'Mean-MAP', 'value': 0.14, 'type': 'SCALAR'}]}}
      #
      #
      # evaluation_measure_result.append(roc_auc_measure)
      # evaluation_measure_result.append(voc_measure)
      # #########################

      self.task_running_statictic[self.ant_name]['measure'] = evaluation_measure_result

    # if self.is_non_mltalker_task:
    #   # generate report resource
    #   logger.info('generate model evaluation report')
    #   everything_to_html(task_running_statictic, os.path.join(self.ant_dump_dir, now_time_stamp))
    #   return

    # significance statistic
    logger.info('Significance difference compare and rank.')
    # benchmark record
    benchmark_model_data = {}
    if self.token is not None:
      response = mlogger.info.benchmark.get()
      if response['status'] == 'OK':
        benchmark_info = response['content']
        for bmd in benchmark_info:
          benchmark_name = bmd['benchmark_name']  # benchmark name (experiment_uuid)
          benchmark_record = bmd['benchmark_record']  # url
          benchmark_report = bmd['benchmark_report']  # 统计数据

          # download benchmark record from url
          logger.info('Download benchmark %s.' % benchmark_name)
          mlogger.info.experiment.download(file_name=benchmark_record,
                                           file_folder=os.path.join(self.ant_dump_dir,
                                                                    self.context.experiment_uuid,
                                                                    'benchmark',
                                                                    benchmark_name),
                                           experiment_uuid=benchmark_name)

          if 'record' not in benchmark_model_data:
            benchmark_model_data['record'] = {}
          benchmark_model_data['record'][benchmark_name] = os.path.join(self.ant_dump_dir,
                                                                        self.context.experiment_uuid,
                                                                        'benchmark',
                                                                        benchmark_name,
                                                                        'record')

          if 'report' not in benchmark_model_data:
            benchmark_model_data['report'] = {}

          for benchmark_experiment_name, benchmark_experiment_report in benchmark_report.items():
            benchmark_model_data['report'][benchmark_name] = benchmark_experiment_report
    elif self.ant_task_benchmark is not None:
      for experiment in self.ant_task_benchmark.split(','):
        if os.path.exists(os.path.join(self.ant_dump_dir, experiment)):
          if 'record' not in benchmark_model_data:
            benchmark_model_data['record'] = {}
          benchmark_model_data['record'][experiment] = os.path.join(self.ant_dump_dir, experiment, 'record')

    if benchmark_model_data is not None and 'record' in benchmark_model_data:
      benchmark_model_record = benchmark_model_data['record']

      self.task_running_statictic[self.ant_name]['significant_diff'] = {}
      for meature_index, measure in enumerate(self.running_task.evaluation_measures):
        if measure.is_support_rank and not measure.crowdsource:
          significant_diff_score = []
          for benchmark_model_name, benchmark_model_address in benchmark_model_record.items():
            if getattr(self.running_task, 'confidence_interval', False):
              with safe_recorder_manager(RecordReader(self.context.recorder.dump_dir)) as record_reader:
                with safe_recorder_manager(RecordReader(benchmark_model_address)) as benchmark_record_reader:
                  s = bootstrap_ab_significance_compare([record_reader, benchmark_record_reader], time.time(), measure,
                                                        10)

                  significant_diff_score.append({'name': benchmark_model_name, 'score': s})
            else:
              compare_value = \
                self.task_running_statictic[self.ant_name]['measure'][meature_index]['statistic']['value'][0]['value'] - \
                benchmark_model_data['report'][benchmark_model_name]['measure'][meature_index]['statistic']['value'][0][
                  'value']
              if compare_value > 0:
                if getattr(measure, 'larger_is_better', 0) == 1:
                  significant_diff_score.append({'name': benchmark_model_name, 'score': 1})
                else:
                  significant_diff_score.append({'name': benchmark_model_name, 'score': -1})
              elif compare_value < 0:
                if getattr(measure, 'larger_is_better', 0) == 1:
                  significant_diff_score.append({'name': benchmark_model_name, 'score': -1})
                else:
                  significant_diff_score.append({'name': benchmark_model_name, 'score': 1})
              else:
                significant_diff_score.append({'name': benchmark_model_name, 'score': 0})

          self.task_running_statictic[self.ant_name]['significant_diff'][measure.name] = significant_diff_score
        elif measure.is_support_rank and measure.crowdsource:
          # TODO: support model significance compare for crowdsource evaluation
          pass

    # error analysis
    logger.info('Error analysis.')
    # benchmark report
    benchmark_model_statistic = None
    if benchmark_model_data is not None and 'report' in benchmark_model_data:
      benchmark_model_statistic = benchmark_model_data['report']

    # task_running_statictic={self.ant_name:
    #                           {'measure':[
    #                             {'statistic': {'name': 'MESR',
    #                                            'value': [{'name': 'MESR', 'value': 0.4, 'type':'SCALAR'}]},
    #                                            'info': [{'id':0,'score':0.8,'category':1},
    #                                                     {'id':1,'score':0.3,'category':1},
    #                                                     {'id':2,'score':0.9,'category':1},
    #                                                     {'id':3,'score':0.5,'category':1},
    #                                                     {'id':4,'score':1.0,'category':1}]},
    #                             {'statistic': {'name': "SE",
    #                                            'value': [{'name': 'SE', 'value': 0.5, 'type': 'SCALAR'}]},
    #                                            'info': [{'id':0,'score':0.4,'category':1},
    #                                                     {'id':1,'score':0.2,'category':1},
    #                                                     {'id':2,'score':0.1,'category':1},
    #                                                     {'id':3,'score':0.5,'category':1},
    #                                                     {'id':4,'score':0.23,'category':1}]}]}}

    for measure_result in self.task_running_statictic[self.ant_name]['measure']:
      if 'info' in measure_result and len(measure_result['info']) > 0:
        measure_name = measure_result['statistic']['name']
        measure_data = measure_result['info']

        # independent analysis per category for classification problem
        measure_data_list = []
        if self.running_task.class_label is not None and len(self.running_task.class_label) > 1:
          if self.running_task.class_label is not None:
            for cl_i, cl in enumerate(self.running_task.class_label):
              measure_data_list.append([md for md in measure_data if md['category'] == cl or md['category'] == cl_i])

        if len(measure_data_list) == 0:
          measure_data_list.append(measure_data)

        for category_id, category_measure_data in enumerate(measure_data_list):
          if len(category_measure_data) == 0:
            continue

          if 'analysis' not in self.task_running_statictic[self.ant_name]:
            self.task_running_statictic[self.ant_name]['analysis'] = {}

          if measure_name not in self.task_running_statictic[self.ant_name]['analysis']:
            self.task_running_statictic[self.ant_name]['analysis'][measure_name] = {}

          # reorganize as list
          method_samples_list = [{'name': self.ant_name, 'data': category_measure_data}]
          if benchmark_model_statistic is not None:
            # extract statistic data from benchmark
            for benchmark_name, benchmark_statistic_data in benchmark_model_statistic.items():
              # finding corresponding measure
              for benchmark_measure_result in benchmark_statistic_data['measure']:
                if benchmark_measure_result['statistic']['name'] == measure_name:
                  benchmark_measure_data = benchmark_measure_result['info']

                  # finding corresponding category
                  sub_benchmark_measure_data = None
                  if self.running_task.class_label is not None and len(self.running_task.class_label) > 1:
                    sub_benchmark_measure_data = \
                      [md for md in benchmark_measure_data if
                       md['category'] == self.running_task.class_label[category_id] or md['category'] == category_id]
                  if sub_benchmark_measure_data is None:
                    sub_benchmark_measure_data = benchmark_measure_data

                  method_samples_list.append({'name': benchmark_name, 'data': sub_benchmark_measure_data})

                  break
              break

          # reorganize data as score matrix
          method_num = len(method_samples_list)
          # samples_num are the same among methods
          samples_num = len(method_samples_list[0]['data'])
          # samples_num = ant_test_dataset.size
          method_measure_mat = np.zeros((method_num, samples_num))
          samples_map = []

          for method_id, method_measure_data in enumerate(method_samples_list):
            # reorder data by index
            order_key = 'id'
            if 'index' in method_measure_data['data'][0]:
              order_key = 'index'
            method_measure_data_order = sorted(method_measure_data['data'], key=lambda x: x[order_key])

            if method_id == 0:
              # record sample id
              for sample_id, sample in enumerate(method_measure_data_order):
                samples_map.append(sample)

            # order consistent
            for sample_id, sample in enumerate(method_measure_data_order):
              method_measure_mat[method_id, sample_id] = sample['score']

          is_binary = False
          # collect all score
          test_score = [td['score'] for td in method_samples_list[0]['data']
                        if td['score'] > -float("inf") and td['score'] < float("inf")]
          hist, x_bins = np.histogram(test_score, 100)
          if len(np.where(hist > 0.0)[0]) <= 2:
            is_binary = True

          # score matrix analysis
          if not is_binary:
            s, ri, ci, lr_samples, mr_samples, hr_samples = \
              continuous_multi_model_measure_analysis(method_measure_mat, samples_map, self.running_dataset)

            analysis_tag = 'Global'
            if len(measure_data_list) > 1:
              analysis_tag = 'Global-Category-' + str(self.running_task.class_label[category_id])

            model_name_ri = [method_samples_list[r]['name'] for r in ri]
            self.task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag] = \
              {'value': s.tolist() if type(s) != list else s,
               'type': 'MATRIX',
               'x': ci,
               'y': model_name_ri,
               'sampling': [{'name': 'High Score Region', 'data': hr_samples},
                            {'name': 'Middle Score Region', 'data': mr_samples},
                            {'name': 'Low Score Region', 'data': lr_samples}]}

            # group by tag
            tags = getattr(self.running_dataset, 'tag', None)
            if tags is not None:
              for tag in tags:
                g_s, g_ri, g_ci, g_lr_samples, g_mr_samples, g_hr_samples = \
                  continuous_multi_model_measure_analysis(method_measure_mat,
                                                          samples_map,
                                                          self.running_dataset,
                                                          filter_tag=tag)

                analysis_tag = 'Group'
                if len(measure_data_list) > 1:
                  analysis_tag = 'Group-Category-' + str(self.running_task.class_label[category_id])

                if analysis_tag not in self.task_running_statictic[self.ant_name]['analysis'][measure_name]:
                  self.task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag] = []

                model_name_ri = [method_samples_list[r]['name'] for r in g_ri]
                tag_data = {'value': g_s.tolist() if type(g_s) != list else g_s,
                            'type': 'MATRIX',
                            'x': g_ci,
                            'y': model_name_ri,
                            'sampling': [{'name': 'High Score Region', 'data': g_hr_samples},
                                         {'name': 'Middle Score Region', 'data': g_mr_samples},
                                         {'name': 'Low Score Region', 'data': g_lr_samples}]}

                self.task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag].append((tag, tag_data))
          else:
            s, ri, ci, region_95, region_52, region_42, region_13, region_one, region_zero = \
              discrete_multi_model_measure_analysis(method_measure_mat,
                                                    samples_map,
                                                    self.running_dataset)

            analysis_tag = 'Global'
            if len(measure_data_list) > 1:
              analysis_tag = 'Global-Category-' + str(self.running_task.class_label[category_id])

            model_name_ri = [method_samples_list[r]['name'] for r in ri]
            self.task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag] = \
              {'value': s.tolist() if type(s) != list else s,
               'type': 'MATRIX',
               'x': ci,
               'y': model_name_ri,
               'sampling': [{'name': '95%', 'data': region_95},
                            {'name': '52%', 'data': region_52},
                            {'name': '42%', 'data': region_42},
                            {'name': '13%', 'data': region_13},
                            {'name': 'best', 'data': region_one},
                            {'name': 'zero', 'data': region_zero}]}

            # group by tag
            tags = getattr(self.running_dataset, 'tag', None)
            if tags is not None:
              for tag in tags:
                g_s, g_ri, g_ci, g_region_95, g_region_52, g_region_42, g_region_13, g_region_one, g_region_zero = \
                  discrete_multi_model_measure_analysis(method_measure_mat,
                                                        samples_map,
                                                        self.running_dataset,
                                                        filter_tag=tag)
                # if 'group' not in task_running_statictic[self.ant_name]['analysis'][measure_name]:
                #   task_running_statictic[self.ant_name]['analysis'][measure_name]['group'] = []
                #
                analysis_tag = 'Group'
                if len(measure_data_list) > 1:
                  analysis_tag = 'Group-Category-' + str(self.running_task.class_label[category_id])

                if analysis_tag not in self.task_running_statictic[self.ant_name]['analysis'][measure_name]:
                  self.task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag] = []

                model_name_ri = [method_samples_list[r]['name'] for r in g_ri]
                tag_data = {'value': g_s.tolist() if type(g_s) != list else g_s,
                            'type': 'MATRIX',
                            'x': g_ci,
                            'y': model_name_ri,
                            'sampling': [{'name': '95%', 'data': region_95},
                                         {'name': '52%', 'data': region_52},
                                         {'name': '42%', 'data': region_42},
                                         {'name': '13%', 'data': region_13},
                                         {'name': 'best', 'data': region_one},
                                         {'name': 'zero', 'data': region_zero}]}

                self.task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag].append((tag, tag_data))

    # 更新实验统计信息
    if self.app_token is not None or self.user_token is not None:
      mlogger.info.experiment.patch(
        experiment_data=zlib.compress(json.dumps({'REPORT': self.task_running_statictic,
                                                  'APP_STAGE': self.stage}).encode()))

    # 生成实验报告
    logger.info('Save experiment report.')
    everything_to_html(self.task_running_statictic,
                       os.path.join(self.ant_dump_dir, self.context.experiment_uuid))

  def start(self):
    # 1.step 加载挑战任务
    running_ant_task = None
    if self.token is not None:
      # 1.1.step 从平台获取挑战任务配置信息
      response = mlogger.info.challenge.get(command=type(self).__name__)
      if response['status'] == 'ERROR':
        logger.error('Couldnt load challenge task.')
        self.token = None
      elif response['status'] == 'SUSPEND':
        # prohibit submit challenge task frequently
        # submit only one in one week
        logger.error('Prohibit submit challenge task frequently.')
        exit(-1)
      elif response['status'] == 'OK':
        content = response['content']

        if 'task' in content:
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('Couldnt load challenge task.')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('Unknow error.')
        exit(-1)

    if running_ant_task is None:
      # 1.2.step 加载自定义任务配置信息
      custom_task = create_task_from_xml(self.ant_task_config, self.context)
      if custom_task is None:
        logger.error('Couldnt load custom task.')
        exit(-1)
      running_ant_task = custom_task

    assert (running_ant_task is not None)
    self._running_task = running_ant_task

    # 3.step  备份实验基本信息
    # 3.1.step 打包代码，并上传至云端
    self.stage = 'CHALLENGE'
    # - backup in dump_dir
    main_folder = self.main_folder
    if not os.path.exists(os.path.join(self.ant_dump_dir, self.context.experiment_uuid)):
      os.makedirs(os.path.join(self.ant_dump_dir, self.context.experiment_uuid))

    # goldcoin = os.path.join(self.ant_dump_dir, experiment_uuid, 'code.tar.gz')

    # if os.path.exists(goldcoin):
    #   os.remove(goldcoin)

    # logger.info('Prepare package model files.')
    # tar = tarfile.open(goldcoin, 'w:gz')

    # # 排除dump目录，将当前文件夹下所有数据上传
    # for root, dirs, files in os.walk(main_folder):
    #   if os.path.commonprefix([root, self.ant_dump_dir]) == self.ant_dump_dir:
    #     continue

    #   rel_root = os.path.relpath(root, main_folder)
    #   for f in files:
    #     tar.add(os.path.join(root, f), arcname=os.path.join(rel_root, f))

    # tar.close()
    # logger.info('Finish package process.')

    # TODO: 在下一个版本支持模型文件上传
    # # 上传模型代码
    # mlogger.getEnv().dashboard.experiment.upload(MODEL=goldcoin,
    #                                              APP_STAGE=self.stage)

    # 3.2.step 更新基本配置
    if self.app_token is not None or self.user_token is not None:
      for k, v in self.context.params.items():
        setattr(mlogger.tag, k, str(v))

    # 4.step 加载测试数据集
    if running_ant_task.dataset_name is None or running_ant_task.dataset_name == '':
      logger.info('Missing dataset set in task config, switch to use --dataset=xxx')
      if self.ant_dataset is None or self.ant_dataset == '':
        logger.error('Sorry, --dataset= is empty.')
        return
      running_ant_task.config(dataset_name=self.ant_dataset)

    logger.info('Loading test dataset %s.'%running_ant_task.dataset_name)

    ant_test_dataset = None
    if self.context.register_at('test') is not None:
      ant_test_dataset = ProxyDataset('test')
      ant_test_dataset.register(test=self.context.register_at('test'))
    else:
      ant_test_dataset = running_ant_task.dataset('test',
                                                  os.path.join(self.ant_data_source, running_ant_task.dataset_name),
                                                  running_ant_task.dataset_params)

    self._running_dataset = ant_test_dataset
    with safe_manager(ant_test_dataset):
      self.context.recorder = RecorderNode2()

      self.stage = "CHALLENGE"
      logger.info('Start infer process.')
      infer_dump_dir = os.path.join(self.ant_dump_dir, self.context.experiment_uuid, 'inference')
      if not os.path.exists(infer_dump_dir):
        os.makedirs(infer_dump_dir)
      else:
        shutil.rmtree(infer_dump_dir)
        os.makedirs(infer_dump_dir)

      if not os.path.exists(os.path.join(infer_dump_dir, 'record')):
        os.makedirs(os.path.join(infer_dump_dir, 'record'))

      intermediate_dump_dir = os.path.join(self.ant_dump_dir, self.context.experiment_uuid, 'record')
      with safe_recorder_manager(self.context.recorder):
        self.context.recorder.dump_dir = intermediate_dump_dir
        # from ablation experiment
        ablation_blocks = getattr(self.ant_context.params, 'ablation', [])
        if ablation_blocks is None:
          ablation_blocks = []
        for b in ablation_blocks:
          self.ant_context.deactivate_block(b)

        if self.context.is_interact_mode:
          logger.info('Running on interact mode.')
          return

        with performance_statistic_region(self.ant_name):
          try:
            self.context.call_infer_process(ant_test_dataset, infer_dump_dir)
          except Exception as e:
            if type(e.__cause__) != StopIteration:
              print(e)
              traceback.print_exc()

        task_running_statictic = get_performance_statistic(self.ant_name)
        self.task_running_statictic = {self.ant_name: task_running_statictic}
        task_running_elapsed_time = self.task_running_statictic[self.ant_name]['time']['elapsed_time']
        self.task_running_statictic[self.ant_name]['time']['elapsed_time_per_sample'] = \
            task_running_elapsed_time / float(ant_test_dataset.size)

        if not self.context.recorder.is_measure:
          # has no annotation to continue to meausre
          # 更新实验统计信息
          if self.app_token is not None or self.user_token is not None:
            mlogger.info.experiment.patch(
              experiment_data=zlib.compress(
                json.dumps({'REPORT': self.task_running_statictic,
                            'APP_STAGE': self.stage
                            }).encode()))

          # 生成实验报告
          logger.info('Save experiment report.')
          everything_to_html(self.task_running_statictic,
                             os.path.join(self.ant_dump_dir, self.context.experiment_uuid))
          return

        # 模型评估
        self.evaluate()


# -*- coding: UTF-8 -*-
# @Time    : 2020/10/26 10:26 上午
# @File    : spider_dataset.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from uuid import uuid4
from antgo.dataflow.dataset.dataset import Dataset
from antgo.ant.utils import *
from antgo.utils import logger
from bs4 import BeautifulSoup
import os
import sys
import cv2
import numpy as nps
import threading
import requests
import re
import uuid
from antgo.ant.download import *
try:
    import queue
except:
    import Queue as queue


class SpiderDataset(Dataset):
  def __init__(self, command_queue=None, dir=None, params=None):
    super(SpiderDataset, self).__init__('test', dir, params)
    self.command_queue = command_queue
    self.waiting_process_queue = queue.Queue()
    self.count = 0
    self.keywords = {}


  def __spider_data_source(self, config):
    datasource_address = config['datasource_address']
    datasource_keywards = config['datasource_keywards']
    dsatasource_params = {}
    if 'datasource_params' in config:
      dsatasource_params = config['datasource_params']
    dsatasource_params.update({
      'download_data_type': 'image'
    })

    if datasource_address not in self.keywords:
      self.keywords[datasource_address] = []
    
    if datasource_keywards in self.keywords[datasource_address]:
      logger.error("Duplicate keywords.")
      return
    self.keywords[datasource_address].append(datasource_keywards)

    if datasource_address == 'baidu':
      baidu_download(datasource_keywards, dsatasource_params, self.dir, self.waiting_process_queue)
    elif datasource_address == 'bing':
      bing_download(datasource_keywards, dsatasource_params, self.dir, self.waiting_process_queue)
    elif datasource_address == 'google':
      google_download(datasource_keywards, dsatasource_params, self.dir, self.waiting_process_queue)
    elif datasource_address == 'vcg':
      vcg_download(datasource_keywards, dsatasource_params, self.dir, self.waiting_process_queue)

  def data_pool(self):
    self.count = 0
    while True:
      try:
        # 1.step 接受爬虫目标指令
        data_pack = self.command_queue.get()
        _, config = data_pack
        self.__spider_data_source(config)

        # 2.step 读取等待处理文件
        image_file = self.waiting_process_queue.get()
        if image_file is None:
          continue

        while True:
          try:
            # read image
            image = cv2.imread(image_file)
            if image is None:
              logger.error("Fail to parse %s"%image_file)
              # get next image file
              image_file = self.waiting_process_queue.get()
              if image_file is None:
                break

              continue
          except:
            logger.error("Fail to parse %s"%image_file)
            # get next image file
            image_file = self.waiting_process_queue.get()
            if image_file is None:
              break

            continue
          
          # increment 1
          self.count += 1

          # return data
          yield image, {}

          # get next image file
          image_file = self.waiting_process_queue.get()
          if image_file is None:
            break
      except:
        logger.error('Fail receive data info.')

  def finish_process_num(self):
    return self.count

  def waiting_process_num(self):
    return self.waiting_process_queue.qsize()

  def at(self, id):
    # 忽略 id
    self.count = 0
    while True:
      try:
        # 1.step 接受爬虫目标指令
        data_pack = self.command_queue.get()
        _, config = data_pack
        self.__spider_data_source(config)

        # 2.step 读取等待处理文件
        image_file = self.waiting_process_queue.get()
        if image_file is None:
          continue

        while True:
          try:
            # read image
            image = cv2.imread(image_file)
            if image is None:
              logger.error("Fail to parse %s" % image_file)
              # get next image file
              image_file = self.waiting_process_queue.get()
              if image_file is None:
                break

              continue
          except:
            logger.error("Fail to parse %s" % image_file)
            # get next image file
            image_file = self.waiting_process_queue.get()
            if image_file is None:
              break

            continue

          # increment 1
          self.count += 1

          # return data
          return image, {}
      except:
        logger.error('Fail receive data info.')

  def split(self, split_params={}, split_method='holdout'):
    raise NotImplemented

  @property
  def size(self):
    return None



# Copyright (c) OpenMMLab. All rights reserved.
from math import sqrt

import torch
import torch.nn.functional as F


def gaussian2D(radius, sigma=1, dtype=torch.float32, device='cpu'):
    """Generate 2D gaussian kernel.

    Args:
        radius (int): Radius of gaussian kernel.
        sigma (int): Sigma of gaussian function. Default: 1.
        dtype (torch.dtype): Dtype of gaussian tensor. Default: torch.float32.
        device (str): Device of gaussian tensor. Default: 'cpu'.

    Returns:
        h (Tensor): Gaussian kernel with a
            ``(2 * radius + 1) * (2 * radius + 1)`` shape.
    """
    x = torch.arange(
        -radius, radius + 1, dtype=dtype, device=device).view(1, -1)
    y = torch.arange(
        -radius, radius + 1, dtype=dtype, device=device).view(-1, 1)

    h = (-(x * x + y * y) / (2 * sigma * sigma)).exp()

    h[h < torch.finfo(h.dtype).eps * h.max()] = 0
    return h


def gen_gaussian_target(heatmap, center, radius, k=1):
    """Generate 2D gaussian heatmap.

    Args:
        heatmap (Tensor): Input heatmap, the gaussian kernel will cover on
            it and maintain the max value.
        center (list[int]): Coord of gaussian kernel's center.
        radius (int): Radius of gaussian kernel.
        k (int): Coefficient of gaussian kernel. Default: 1.

    Returns:
        out_heatmap (Tensor): Updated heatmap covered by gaussian kernel.
    """
    diameter = 2 * radius + 1
    gaussian_kernel = gaussian2D(
        radius, sigma=diameter / 6, dtype=heatmap.dtype, device=heatmap.device)

    x, y = center

    height, width = heatmap.shape[:2]

    left, right = min(x, radius), min(width - x, radius + 1)
    top, bottom = min(y, radius), min(height - y, radius + 1)

    masked_heatmap = heatmap[y - top:y + bottom, x - left:x + right]
    masked_gaussian = gaussian_kernel[radius - top:radius + bottom,
                                      radius - left:radius + right]
    out_heatmap = heatmap
    torch.max(
        masked_heatmap,
        masked_gaussian * k,
        out=out_heatmap[y - top:y + bottom, x - left:x + right])

    return out_heatmap


def gaussian_radius(det_size, min_overlap):
    r"""Generate 2D gaussian radius.

    This function is modified from the `official github repo
    <https://github.com/princeton-vl/CornerNet-Lite/blob/master/core/sample/
    utils.py#L65>`_.

    Given ``min_overlap``, radius could computed by a quadratic equation
    according to Vieta's formulas.

    There are 3 cases for computing gaussian radius, details are following:

    - Explanation of figure: ``lt`` and ``br`` indicates the left-top and
      bottom-right corner of ground truth box. ``x`` indicates the
      generated corner at the limited position when ``radius=r``.

    - Case1: one corner is inside the gt box and the other is outside.

    .. code:: text

        |<   width   >|

        lt-+----------+         -
        |  |          |         ^
        +--x----------+--+
        |  |          |  |
        |  |          |  |    height
        |  | overlap  |  |
        |  |          |  |
        |  |          |  |      v
        +--+---------br--+      -
           |          |  |
           +----------+--x

    To ensure IoU of generated box and gt box is larger than ``min_overlap``:

    .. math::
        \cfrac{(w-r)*(h-r)}{w*h+(w+h)r-r^2} \ge {iou} \quad\Rightarrow\quad
        {r^2-(w+h)r+\cfrac{1-iou}{1+iou}*w*h} \ge 0 \\
        {a} = 1,\quad{b} = {-(w+h)},\quad{c} = {\cfrac{1-iou}{1+iou}*w*h}
        {r} \le \cfrac{-b-\sqrt{b^2-4*a*c}}{2*a}

    - Case2: both two corners are inside the gt box.

    .. code:: text

        |<   width   >|

        lt-+----------+         -
        |  |          |         ^
        +--x-------+  |
        |  |       |  |
        |  |overlap|  |       height
        |  |       |  |
        |  +-------x--+
        |          |  |         v
        +----------+-br         -

    To ensure IoU of generated box and gt box is larger than ``min_overlap``:

    .. math::
        \cfrac{(w-2*r)*(h-2*r)}{w*h} \ge {iou} \quad\Rightarrow\quad
        {4r^2-2(w+h)r+(1-iou)*w*h} \ge 0 \\
        {a} = 4,\quad {b} = {-2(w+h)},\quad {c} = {(1-iou)*w*h}
        {r} \le \cfrac{-b-\sqrt{b^2-4*a*c}}{2*a}

    - Case3: both two corners are outside the gt box.

    .. code:: text

           |<   width   >|

        x--+----------------+
        |  |                |
        +-lt-------------+  |   -
        |  |             |  |   ^
        |  |             |  |
        |  |   overlap   |  | height
        |  |             |  |
        |  |             |  |   v
        |  +------------br--+   -
        |                |  |
        +----------------+--x

    To ensure IoU of generated box and gt box is larger than ``min_overlap``:

    .. math::
        \cfrac{w*h}{(w+2*r)*(h+2*r)} \ge {iou} \quad\Rightarrow\quad
        {4*iou*r^2+2*iou*(w+h)r+(iou-1)*w*h} \le 0 \\
        {a} = {4*iou},\quad {b} = {2*iou*(w+h)},\quad {c} = {(iou-1)*w*h} \\
        {r} \le \cfrac{-b+\sqrt{b^2-4*a*c}}{2*a}

    Args:
        det_size (list[int]): Shape of object.
        min_overlap (float): Min IoU with ground truth for boxes generated by
            keypoints inside the gaussian kernel.

    Returns:
        radius (int): Radius of gaussian kernel.
    """
    height, width = det_size

    a1 = 1
    b1 = (height + width)
    c1 = width * height * (1 - min_overlap) / (1 + min_overlap)
    sq1 = sqrt(b1**2 - 4 * a1 * c1)
    r1 = (b1 - sq1) / (2 * a1)

    a2 = 4
    b2 = 2 * (height + width)
    c2 = (1 - min_overlap) * width * height
    sq2 = sqrt(b2**2 - 4 * a2 * c2)
    r2 = (b2 - sq2) / (2 * a2)

    a3 = 4 * min_overlap
    b3 = -2 * min_overlap * (height + width)
    c3 = (min_overlap - 1) * width * height
    sq3 = sqrt(b3**2 - 4 * a3 * c3)
    r3 = (b3 + sq3) / (2 * a3)
    return min(r1, r2, r3)


def get_local_maximum(heat, kernel=3):
    """Extract local maximum pixel with given kernel.

    Args:
        heat (Tensor): Target heatmap.
        kernel (int): Kernel size of max pooling. Default: 3.

    Returns:
        heat (Tensor): A heatmap where local maximum pixels maintain its
            own value and other positions are 0.
    """
    pad = (kernel - 1) // 2
    hmax = F.max_pool2d(heat, kernel, stride=1, padding=pad)
    keep = (hmax == heat).float()
    return heat * keep


def get_topk_from_heatmap(scores, k=20):
    """Get top k positions from heatmap.

    Args:
        scores (Tensor): Target heatmap with shape
            [batch, num_classes, height, width].
        k (int): Target number. Default: 20.

    Returns:
        tuple[torch.Tensor]: Scores, indexes, categories and coords of
            topk keypoint. Containing following Tensors:

        - topk_scores (Tensor): Max scores of each topk keypoint.
        - topk_inds (Tensor): Indexes of each topk keypoint.
        - topk_clses (Tensor): Categories of each topk keypoint.
        - topk_ys (Tensor): Y-coord of each topk keypoint.
        - topk_xs (Tensor): X-coord of each topk keypoint.
    """
    batch, _, height, width = scores.size()
    topk_scores, topk_inds = torch.topk(scores.view(batch, -1), k)
    topk_clses = topk_inds // (height * width)
    topk_inds = topk_inds % (height * width)
    topk_ys = topk_inds // width
    topk_xs = (topk_inds % width).int().float()
    return topk_scores, topk_inds, topk_clses, topk_ys, topk_xs


def gather_feat(feat, ind, mask=None):
    """Gather feature according to index.

    Args:
        feat (Tensor): Target feature map.
        ind (Tensor): Target coord index.
        mask (Tensor | None): Mask of feature map. Default: None.

    Returns:
        feat (Tensor): Gathered feature.
    """
    dim = feat.size(2)
    ind = ind.unsqueeze(2).repeat(1, 1, dim)
    feat = feat.gather(1, ind)
    if mask is not None:
        mask = mask.unsqueeze(2).expand_as(feat)
        feat = feat[mask]
        feat = feat.view(-1, dim)
    return feat


def transpose_and_gather_feat(feat, ind):
    """Transpose and gather feature according to index.

    Args:
        feat (Tensor): Target feature map.
        ind (Tensor): Target coord index.

    Returns:
        feat (Tensor): Transposed and gathered feature.
    """
    feat = feat.permute(0, 2, 3, 1).contiguous()
    feat = feat.view(feat.size(0), -1, feat.size(3))
    feat = gather_feat(feat, ind)
    return feat



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 21:56
# @File    : deploy.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
import os
import json
from antgo.pipeline.functional.common.config import *
from antgo.pipeline import extent
from antgo.pipeline.extent.glue.common import *
from antgo.pipeline.extent.op.loader import *
import onnx
import onnxruntime
import re
import shutil
import subprocess


def snpe_import_config(output_folder, project_name, platform, abi, device='GPU'):
    # load snpe lib
    # step1: 下载snpe库, 并解压到固定为止
    root_folder = os.path.abspath('.3rd/')
    os.makedirs(root_folder, exist_ok=True)
    if not os.path.exists(os.path.join(root_folder, 'snpe-2.9.0.4462')):
        os.system(f'cd .3rd/ && wget http://experiment.mltalker.com/snpe-2.9.0.4462.zip && unzip snpe-2.9.0.4462.zip')
    snpe_path = os.path.join(root_folder, 'snpe-2.9.0.4462')

    # step2: 推送依赖库到包位置
    os.makedirs(os.path.join(output_folder, '3rd', abi), exist_ok=True)
    shutil.copyfile(os.path.join(f'{snpe_path}/lib/aarch64-android-clang8.0', 'libSNPE.so'), os.path.join(output_folder, '3rd', abi, 'libSNPE.so'))

    # step3: 生成cmake代码片段
    snpe_cmake_code_snippet = ''
    snpe_cmake_code_snippet += f'set(SNPE_LIB_DIR {snpe_path}/lib/aarch64-android-clang8.0)\n'
    snpe_cmake_code_snippet += f'set(SNPE_INCLUDE_DIR {snpe_path}/include/zdl)\n'
    snpe_cmake_code_snippet += 'include_directories(${SNPE_INCLUDE_DIR})\n'
    snpe_cmake_code_snippet += 'add_library(libSNPE SHARED IMPORTED)\n'
    snpe_cmake_code_snippet += 'set_target_properties(libSNPE PROPERTIES IMPORTED_LOCATION ${SNPE_LIB_DIR}/libSNPE.so)\n'
    snpe_cmake_code_snippet += f'target_link_libraries({project_name} libSNPE)\n'

    code_line_list = []
    for line in open(os.path.join(output_folder, 'CMakeLists.txt')):
        if len(code_line_list) > 0 and code_line_list[-1].strip() == '# model engine' and line == '\n':
            code_line_list.append(snpe_cmake_code_snippet)

        code_line_list.append(line)

    with open(os.path.join(output_folder, 'CMakeLists.txt'), 'w') as fp:
        for line in code_line_list:
            fp.write(line)


def rknn_import_config(output_folder, project_name, platform, abi, device='rk3588'):
    # step1: 下载rknn库，并解压到固定为止
    root_folder = os.path.abspath('.3rd/')
    os.makedirs(root_folder, exist_ok=True)
    if not os.path.exists(os.path.join(root_folder, 'rknpu2')):
        os.system(f'cd .3rd/ && git clone https://github.com/rockchip-linux/rknpu2.git')
    rknn_path = os.path.join(root_folder, 'rknpu2')

    rknn_runtime_folder = os.path.join(rknn_path, 'runtime')
    assert(os.path.exists(os.path.join(rknn_runtime_folder, device.upper())))
    device_rknn_runtime_folder = os.path.join(rknn_runtime_folder, device.upper())
    assert(os.path.exists(os.path.join(device_rknn_runtime_folder, platform.capitalize())))
    RKNN_API_PATH = os.path.join(device_rknn_runtime_folder, platform.capitalize(), 'librknn_api')
    # device_rknn_so_folder = os.path.join(device_rknn_runtime_folder, platform.capitalize(), 'librknn_api', abi)
    # device_rknn_include_folder = os.path.join(device_rknn_runtime_folder, platform.capitalize(), 'librknn_api', 'include')

    # step2: 推送依赖库到包为止
    os.makedirs(os.path.join(output_folder, '3rd', abi), exist_ok=True)
    shutil.copyfile(os.path.join(RKNN_API_PATH, abi, 'librknnrt.so'), os.path.join(output_folder, '3rd', abi, 'librknnrt.so'))

    # step3: 生成cmake代码片段
    rknn_cmake_code_snippet = f'set(RKNN_API_PATH {RKNN_API_PATH})\n'
    rknn_cmake_code_snippet += f'include_directories({RKNN_API_PATH}/include)\n'
    rknn_cmake_code_snippet += f'set(RKNN_RT_LIB {RKNN_API_PATH}/{abi}/librknnrt.so)\n'
    rknn_cmake_code_snippet += 'add_library(librknnrt SHARED IMPORTED)\n'
    rknn_cmake_code_snippet += 'set_target_properties(librknnrt PROPERTIES IMPORTED_LOCATION ${RKNN_RT_LIB})\n'
    rknn_cmake_code_snippet += f'target_link_libraries({project_name} librknnrt)\n'

    code_line_list = []
    for line in open(os.path.join(output_folder, 'CMakeLists.txt')):
        if len(code_line_list) > 0 and code_line_list[-1].strip() == '# model engine' and line == '\n':
            code_line_list.append(rknn_cmake_code_snippet)

        code_line_list.append(line)

    with open(os.path.join(output_folder, 'CMakeLists.txt'), 'w') as fp:
        for line in code_line_list:
            fp.write(line)


def tensorrt_import_config(output_folder, project_name, platform, abi, device=''):
    # step1: 下载tensorrt库，并解压到固定为止
    root_folder = os.path.abspath('.3rd/')
    os.makedirs(root_folder, exist_ok=True)
    
    # tensorrt 仅支持linux
    assert(platform.lower() == 'linux')

    # 涉及cudnn, tensorrt, cuda库
    # 下载 cudnn cudnn-linux-x86_64-8.8.0.121_cuda12-archive.tar.xz https://developer.nvidia.com/rdp/cudnn-archive
    # 下载 tensorrt TensorRT-8.6.1.6 https://developer.nvidia.com/nvidia-tensorrt-8x-download
    if not os.path.exists(os.path.join(root_folder, 'cudnn-linux-x86_64-8.8.0.121_cuda12-archive')):
        print('download cudnn')
        os.system(f'cd .3rd/ && wget http://experiment.mltalker.com/cudnn-linux-x86_64-8.8.0.121_cuda12-archive.tar.xz && tar -xf cudnn-linux-x86_64-8.8.0.121_cuda12-archive.tar.xz')
    if not os.path.exists(os.path.join(root_folder, 'TensorRT-8.6.1.6')):
        print('download tensorrt')
        os.system(f'cd .3rd/ && wget http://experiment.mltalker.com/TensorRT-8.6.1.6.Linux.x86_64-gnu.cuda-12.0.tar.gz && tar -xf TensorRT-8.6.1.6.Linux.x86_64-gnu.cuda-12.0.tar.gz')

    cudnn_path = os.path.join(root_folder, 'cudnn-linux-x86_64-8.8.0.121_cuda12-archive')
    tensorrt_path = os.path.join(root_folder, 'TensorRT-8.6.1.6')

    # step2: 推送依赖库到包为止
    os.makedirs(os.path.join(output_folder, '3rd', abi), exist_ok=True)

    # step3: 生成cmake代码片段
    tensorrt_cmake_code_snippet = f'set(TensorRT_DIR {tensorrt_path})\n'
    tensorrt_cmake_code_snippet += 'set(CUDA_TOOLKIT_ROOT_DIR /usr/local/cuda)\n'
    tensorrt_cmake_code_snippet += 'find_package(TensorRT REQUIRED)\n'
    tensorrt_cmake_code_snippet += 'find_package(CUDA REQUIRED)\n'
    tensorrt_cmake_code_snippet += f'target_include_directories({project_name} '+'PUBLIC ${CUDA_INCLUDE_DIRS} ${TensorRT_INCLUDE_DIRS})\n'
    tensorrt_cmake_code_snippet += f'target_link_libraries({project_name} '+'${CUDA_LIBRARIES} ${CMAKE_THREAD_LIBS_INIT} ${TensorRT_LIBRARIES})\n'

    code_line_list = []
    for line in open(os.path.join(output_folder, 'CMakeLists.txt')):
        if len(code_line_list) > 0 and code_line_list[-1].strip() == '# model engine' and line == '\n':
            code_line_list.append(tensorrt_cmake_code_snippet)

        code_line_list.append(line)

    with open(os.path.join(output_folder, 'CMakeLists.txt'), 'w') as fp:
        for line in code_line_list:
            fp.write(line)

    # step4: 更新run.sh片段
    code_line_list = []
    for line in open(os.path.join(output_folder, 'run.sh')):
        if 'export LD_LIBRARY_PATH=' in line:
            aa,bb,cc = line.split(';')
            bb = f'export LD_LIBRARY_PATH=.:{cudnn_path}/lib:{tensorrt_path}/lib'
            line = f'{aa}; {bb}; {cc}'
        code_line_list.append(line)
    with open(os.path.join(output_folder, 'run.sh'), 'w') as fp:
        for line in code_line_list:
            fp.write(line)


def auto_generate_eagleeye_op(op_name, op_index, op_args, op_kwargs, output_folder):
    func = getattr(extent.func, op_name)

    input_ctx, output_ctx = op_index
    if isinstance(input_ctx, str):
        input_ctx = [input_ctx]
    if isinstance(output_ctx, str):
        output_ctx = [output_ctx]
    
    # 创建header文件
    eagleeye_warp_h_code_content = \
        gen_code('./templates/op_code.h')(            
            op_name=f"{op_name.replace('_','').capitalize()}Op",
            input_num=len(input_ctx),
            output_num=len(output_ctx)
        )
    
    # folder tree
    # output_folder/
    #   include/
    #   src/
    include_folder = os.path.join(output_folder, 'extent','include')
    os.makedirs(include_folder, exist_ok=True)
    with open(os.path.join(include_folder, f'{op_name}_op_warp.h'), 'w') as fp:
        fp.write(eagleeye_warp_h_code_content)

    # 创建cpp文件
    # 函数参数部分
    func_args = [None] * len(func.func.arg_names)
    for i, (var_name, var_type) in enumerate(zip(func.func.arg_names, func.func.arg_types)):
        if i < len(op_args):
            func_args[i] =  ('param', var_name, var_type, op_args[i])

    for i, (var_name, var_type) in enumerate(zip(func.func.arg_names, func.func.arg_types)):
        if op_kwargs.get(var_name, None) is not None and var_type.is_const:
            func_args[i] = ('param', var_name, var_type, op_kwargs.get(var_name))

    # 函数输入部分
    input_i = 0
    for i, (var_name, var_type) in enumerate(zip(func.func.arg_names, func.func.arg_types)):
        if var_type.is_const and func_args[i] is None:
            func_args[i] = (f'input_{input_i}',var_name, var_type, None)
            input_i += 1

    # 函数输出部分
    output_i = 0
    for i, (var_name, var_type) in enumerate(zip(func.func.arg_names, func.func.arg_types)):
        if not var_type.is_const and func_args[i] is None:
            func_args[i] = (f'output_{output_i}', var_name, var_type, None)
            output_i += 1
    
    # 从Tensor到C*Tensor转换
    convert_map_1 = {
        'CFTensor': 'convert_tensor_cftensor',
        'CITensor': 'convert_tensor_citensor',
        'CUCTensor':'convert_tensor_cuctensor',
        'CDTensor': 'convert_tensor_cdtensor',
    }
    # 从C*Tensor到Tensor转换
    convert_map_2 = {
        'CFTensor': 'convert_cftensor_tensor',
        'CITensor': 'convert_citensor_tensor',
        'CUCTensor':'convert_cuctensor_tensor',
        'CDTensor': 'convert_cdtensor_tensor',
    }
    
    # 创建C*Tensor    
    new_map = {
        'CFTensor': 'new_cftensor',
        'CITensor': 'new_citensor',
        'CUCTensor':'new_cuctensor',
        'CDTensor': 'new_cdtensor',
    }
    
    # 初始化C*Tensor
    init_map = {
        '<f4': 'init_cftensor',
        '<i4': 'init_citensor',
        '|u1': 'init_cuctensor'
    }
    
    args_convert = ''
    output_covert = ''
    args_clear = ''
    for arg in func_args:
        flag, arg_name, arg_type, arg_value = arg
        if flag.startswith('input'):
            input_p = int(flag[6:])
            arg_type = arg_type.cname.replace('*','')
            args_convert += f'{arg_type}*{arg_name}={convert_map_1[arg_type]}(input[{input_p}]);\n'
            
            args_clear += f'{arg_name}->destroy();\ndelete {arg_name};\n'
        elif flag.startswith('output'):
            arg_type = arg_type.cname.replace('*','')
            args_convert += f'{arg_type}*{arg_name}={new_map[arg_type]}();\n'
            
            output_p = int(flag[7:])
            output_covert += f'm_outputs[{output_p}]={convert_map_2[arg_type]}({arg_name});\n'
            args_clear += f'{arg_name}->destroy();\ndelete {arg_name};\n'
        else:
            if isinstance(arg_value, np.ndarray):
                assert(arg_value.dtype == np.float32 or arg_value.dtype == np.int32 or arg_value.dtype == np.uint8)
                
                data_value_str = '{'+','.join([f'{v}' for v in arg_value.tolist()])+'}'
                data_shape_str = '{'+','.join([f'{v}' for v in arg_value.shape])+'}'
                arg_type = arg_type.cname.replace('*','')
                args_convert += f'{arg_type}*{arg_name}={init_map[arg_value.dtype.str]}({data_value_str}, {data_shape_str});\n'
                
                args_clear += f'{arg_name}->destroy();\ndelete {arg_name};\n'
            else:
                arg_type = arg_type.cname.replace('const', '')
                args_convert += f'{arg_type} {arg_name}={str(arg_value).lower()};\n'
           
    args_inst = ''
    for _, arg_name, _, _ in func_args:
        if args_inst == '':
            args_inst = f'{arg_name}'
        else:
            args_inst += f',{arg_name}'
    
    eagleeye_warp_cpp_code_content = \
        gen_code('./templates/op_code.cpp')(
            op_name=f"{op_name.replace('_','').capitalize()}Op",
            func_name=op_name,
            inc_fname1=os.path.abspath(os.path.join(output_folder, 'extent', 'include', f'{op_name}_op_warp.h')),
            inc_fname2=os.path.abspath(func.func.loader_kwargs['cpp_info'].cpp_fname),
            args_convert=args_convert,
            args_inst=args_inst,
            return_statement='',
            output_covert=output_covert,
            args_clear=args_clear
        )

    src_folder = os.path.join(output_folder, 'extent', 'src')
    os.makedirs(src_folder, exist_ok=True)
    with open(os.path.join(src_folder, f'{op_name}_op_warp.cpp'), 'w') as fp:
        fp.write(eagleeye_warp_cpp_code_content)

    info = {
        'type': f"{op_name.replace('_','').capitalize()}Op",
        'input': input_ctx,
        'output': output_ctx,
        'args': {},
        'include': os.path.join('extent','include', f'{op_name}_op_warp.h'),
        'src': os.path.join('./', 'extent', 'src', f'{op_name}_op_warp.cpp')
    }
    return info


def convert_args_eagleeye_op_args(op_args, op_kwargs):
    # ignore op_args
    converted_op_args = {}
    for arg_name, arg_info in op_kwargs.items():
        if isinstance(arg_info, np.ndarray):
            # numpy
            converted_op_args[arg_name] = arg_info.flatten().astype(np.float32)
        elif isinstance(arg_info, list) or isinstance(arg_info, tuple):
            # list
            converted_op_args[arg_name] = np.array(arg_info).flatten().astype(np.float32)
        else:
            # scalar
            converted_op_args[arg_name] = [float(arg_info)]

    converted_op_args['c++_type'] = 'std::map<std::string, std::vector<float>>'
    return converted_op_args


def update_cmakelist(output_folder, project_name, src_op_warp_list):
    info = []
    is_found_include_directories_insert = False
    is_start_add_src_code = False

    src_op_warp_flag = [0 for _ in range(len(src_op_warp_list))]
    for line in open(os.path.join(output_folder, 'CMakeLists.txt')):
        if is_start_add_src_code and not line.strip().endswith(')'):
            for src_op_warp_file_i, src_op_warp_file in enumerate(src_op_warp_list):
                if src_op_warp_file == line.strip():
                    src_op_warp_flag[src_op_warp_file_i] = 1
                    break
        if is_start_add_src_code and line.strip().endswith(')'):
            for src_op_warp_file_i, src_op_warp_file_flag in enumerate(src_op_warp_flag):
                if src_op_warp_file_flag == 0:
                    info.append(f'{src_op_warp_list[src_op_warp_file_i]}\n')

            is_start_add_src_code = False

        if f'set({project_name}_SRC' in line:
            is_start_add_src_code = True

        if line.startswith('include_directories') and not is_found_include_directories_insert:
            extent_cpp_include_folder = os.path.dirname(os.path.realpath(__file__))
            extent_cpp_include_folder = os.path.dirname(extent_cpp_include_folder)
            extent_cpp_include_folder = os.path.dirname(extent_cpp_include_folder)
            extent_cpp_include_folder = os.path.join(extent_cpp_include_folder, 'extent','cpp','include')
            if extent_cpp_include_folder not in line:
                info.append(f'include_directories({extent_cpp_include_folder})\n')

            is_found_include_directories_insert = True

        info.append(line)
    
    with open(os.path.join(output_folder, 'CMakeLists.txt'), 'w') as fp:
        for line in info:
            fp.write(line)


tensortype_map = {
    'tensor(float)': 6
}


def convert_onnx_to_platform_engine(op_name, op_index, op_args, op_kwargs, output_folder, platform, abi, project_name):
    # 0.step 参数
    platform_engine = op_kwargs.get('engine', 'tensorrt')   # 在平台，使用什么模型引擎
    platform_engine_args = op_kwargs.get('engine_args', {}) # 在平台，使用什么模型引擎参数
    platform_device = platform_engine_args.get('device', 'CPU')        # 在平台，使用什么设备

    # 1.step 转换模型格式文件
    # TODO, 临时
    platform_model_path = platform_engine_args.get('model', None)
    if platform_model_path is None and platform == 'android':
        onnx_file_path = op_kwargs.get('onnx_path')
        # TODO,支持自动转换模型
        if platform_engine == 'snpe':
            if platform_engine_args.get('quantize', False):
                # 转量化模型
                os.system(f'mkdir /tmp/onnx && mkdir /tmp/onnx/snpe && cp {onnx_file_path} /tmp/onnx/')
                # 确保存在校正数据集
                assert(os.path.exists(platform_engine_args.get('calibration-images')))
                shutil.copytree(platform_engine_args.get('calibration-images'), '/tmp/onnx/calibration-images')

                prefix = os.path.basename(onnx_file_path)[:-5]
                onnx_dir_path = os.path.dirname(onnx_file_path)
                os.system(f'cd /tmp/onnx && docker run -v $(pwd):/workspace snpeconvert bash convert.sh --i={prefix}.onnx --o=./snpe/{prefix} --quantize --npu --data-folder=calibration-images')
                converted_model_file = ''
                for file_name in os.listdir('/tmp/onnx/snpe/'):
                    if file_name[0] != '.':
                        converted_model_file = file_name
                        break

                os.system(f'cp -r /tmp/onnx/snpe/* {onnx_dir_path} && rm -rf /tmp/onnx/')
                platform_model_path = os.path.join(onnx_dir_path, converted_model_file)
            else:
                # 转浮点模型
                os.system(f'mkdir /tmp/onnx && mkdir /tmp/onnx/snpe && cp {onnx_file_path} /tmp/onnx/')

                prefix = os.path.basename(onnx_file_path)[:-5]
                onnx_dir_path = os.path.dirname(onnx_file_path)
                os.system(f'cd /tmp/onnx && docker run -v $(pwd):/workspace snpeconvert bash convert.sh --i={prefix}.onnx --o=./snpe/{prefix}')
                converted_model_file = ''
                for file_name in os.listdir('/tmp/onnx/snpe/'):
                    if file_name[0] != '.':
                        converted_model_file = file_name
                        break

                os.system(f'cp -r /tmp/onnx/snpe/* {onnx_dir_path} && rm -rf /tmp/onnx/')
                platform_model_path = os.path.join(onnx_dir_path, converted_model_file)
        elif platform_engine == 'rknn':
            if platform_engine_args.get('quantize', False):
                # 转量化模型
                os.system(f'mkdir /tmp/onnx && mkdir /tmp/onnx/rknn && cp {onnx_file_path} /tmp/onnx/')
                # 确保存在校正数据集
                assert(os.path.exists(platform_engine_args.get('calibration-images')))
                shutil.copytree(platform_engine_args.get('calibration-images'), '/tmp/onnx/calibration-images')

                prefix = os.path.basename(onnx_file_path)[:-5]
                onnx_dir_path = os.path.dirname(onnx_file_path)
                mean_values = ','.join([str(v) for v in  op_kwargs.get('mean')])
                std_values = ','.join([str(v) for v in  op_kwargs.get('std')])
                os.system(f'cd /tmp/onnx && docker run -v $(pwd):/workspace rknnconvert bash convert.sh --i={prefix}.onnx --o=./rknn/{prefix} --device={platform_device} --mean-values={mean_values} --std-values={std_values}')
                converted_model_file = ''
                for file_name in os.listdir('/tmp/onnx/rknn/'):
                    if file_name[0] != '.':
                        converted_model_file = file_name
                        break

                os.system(f'cp -r /tmp/onnx/rknn/* {onnx_dir_path} && rm -rf /tmp/onnx/')
                platform_model_path = os.path.join(onnx_dir_path, converted_model_file)
            else:
                # 转浮点模型
                os.system(f'mkdir /tmp/onnx && mkdir /tmp/onnx/rknn && cp {onnx_file_path} /tmp/onnx/')
                
                prefix = os.path.basename(onnx_file_path)[:-5]
                onnx_dir_path = os.path.dirname(onnx_file_path)
                mean_values = ','.join([str(v) for v in  op_kwargs.get('mean')])
                std_values = ','.join([str(v) for v in  op_kwargs.get('std')])
                os.system(f'cd /tmp/onnx && docker run -v $(pwd):/workspace rknnconvert bash convert.sh --i={prefix}.onnx --o=./rknn/{prefix} --device={platform_device} --mean-values={mean_values} --std-values={std_values}')
                converted_model_file = ''
                for file_name in os.listdir('/tmp/onnx/rknn/'):
                    if file_name[0] != '.':
                        converted_model_file = file_name
                        break
                
                os.system(f'cp -r /tmp/onnx/rknn/* {onnx_dir_path} && rm -rf /tmp/onnx/')
                platform_model_path = os.path.join(onnx_dir_path, converted_model_file)
        elif platform_engine == 'tnn':
            print('TODO support tnn')
            pass

    if platform_model_path is None and platform == 'linux':
        platform_model_path = op_kwargs.get('onnx_path')

    # 将平台关联模型放入输出目录中
    if os.path.exists(platform_model_path):
        os.makedirs(os.path.join(output_folder, 'model'), exist_ok=True)
        shutil.copyfile(platform_model_path, os.path.join(output_folder, 'model', os.path.basename(platform_model_path)))

    # 2.step 参数转换及代码生成
    config_func_map = {
        'snpe': snpe_import_config,
        'rknn': rknn_import_config,
        'tensorrt': tensorrt_import_config
    }
    config_func_map[platform_engine](output_folder, project_name, platform, abi, platform_device)

    input_ctx, output_ctx = op_index
    if not isinstance(input_ctx, tuple):
        input_ctx = (input_ctx,)
    if not isinstance(output_ctx, tuple):
        output_ctx = (output_ctx,)

    template_args = f'<{len(input_ctx)},{len(output_ctx)}>'

    # 解析onnx模型获得输入、输出信息
    onnx_session = onnxruntime.InferenceSession(op_kwargs['onnx_path'], providers=['CPUExecutionProvider'])
    mode_name = os.path.basename(op_kwargs['onnx_path']).split('.')[0]
    input_names = []
    input_shapes = []
    input_types = []
    output_names = []
    output_shapes = []
    output_types = []
    model_folder = f'/sdcard/{project_name}/.model/' if platform == 'android' else os.path.dirname(op_kwargs['onnx_path'])     # 考虑将转好的模型放置的位置
    writable_path = f'/sdcard/{project_name}/.tmp/' if platform == 'android' else os.path.dirname(op_kwargs['onnx_path'])        # 考虑到 设备可写权限位置(android)

    # 更新run.sh（仅设备端运行时需要添加推送模型代码）
    if platform.lower() == 'android':
        run_shell_code_list = []
        is_found_model_push_line = False
        is_found_model_platform_folder = False
        for line in open(os.path.join(output_folder, 'run.sh')):
            if line.startswith(f'adb push {platform_model_path}'):
                # 替换
                line = f'adb push {platform_model_path} {model_folder}\n'
                is_found_model_push_line = True

            if f'mkdir -p {model_folder};' in line:
                is_found_model_platform_folder = True

            if line.startswith('adb shell "cd /data/local/tmp') and not is_found_model_platform_folder:
                # 插入
                run_shell_code_list.append(f'adb shell "if [ ! -d {model_folder} ]; then mkdir -p {model_folder}; fi;"\n')

            if line.startswith('adb shell "cd /data/local/tmp') and not is_found_model_push_line:
                # 插入
                run_shell_code_list.append(f'adb push {platform_model_path} {model_folder}\n')
            

            run_shell_code_list.append(line)

        with open(os.path.join(output_folder, 'run.sh'), 'w') as fp:
            for line in run_shell_code_list:
                fp.write(line)

    num_threads = 2
    for input_tensor in onnx_session.get_inputs():
        input_names.append(input_tensor.name)
        input_shapes.append(input_tensor.shape)
        input_types.append(tensortype_map[input_tensor.type])

    alias_output_names = []
    for output_tensor in onnx_session.get_outputs():
        if 'alias_output_names' in platform_engine_args:
            output_names.append(platform_engine_args['alias_output_names'][output_tensor.name])
            alias_output_names.append(output_tensor.name)
        else:
            output_names.append(output_tensor.name)
        output_shapes.append(output_tensor.shape)
        output_types.append(tensortype_map[output_tensor.type])

    # string args, vector args, other args
    model_name = '.'.join(platform_model_path.split("/")[-1].split('.')[:-1])
    op_args = (
        {
            'model_name': [f'"{model_name}"'],
            'device': [f'"{platform_device}"'],
            'input_names': [f'"{c}"' for c in input_names],
            'output_names': [f'"{c}"' for c in output_names],
            'alias_output_names': [f'"{c}"' for c in alias_output_names],
            'model_folder': [f'"{model_folder}"'],
            'writable_path': [f'"{writable_path}"'],
            'c++_type': 'std::map<std::string, std::vector<std::string>>'
        },
        {
            'input_shapes': ['{'+','.join(str(m) for m in n)+'}' for n in input_shapes],
            'output_shapes': ['{'+','.join(str(m) for m in n)+'}' for n in output_shapes],
            'c++_type': 'std::map<std::string, std::vector<std::vector<float>>>'
        },
        {
            'input_types': input_types,
            'output_types': output_types,
            'num_threads': [num_threads],
            'c++_type': 'std::map<std::string, std::vector<float>>'
        }
    )

    if op_kwargs.get('mean', None) is not None:
        op_args[-1]['mean'] = ['{'+','.join(str(m) for m in op_kwargs['mean'])+'}']
        op_args[-1]['std'] = ['{'+','.join(str(m) for m in op_kwargs['std'])+'}']
        op_args[-1]['rgb2bgr'] = ['{1}'] if op_kwargs.get('rgb2bgr', False) else ['{0}']

    include_path = f'eagleeye/engine/nano/op/{platform_engine.lower()}_op.hpp'
    return f'{platform_engine.capitalize()}Op', template_args, op_args, include_path


def package_build(output_folder, eagleeye_path, project_config, platform, abi=None):    
    # 获得eagleeye核心算子集合
    core_op_set = load_eagleeye_op_set(eagleeye_path)

    # 获得计算图配置信息
    graph_config = get_graph_info()['op']

    # 准备算子函数代码
    deploy_graph_info = {}
    op_name_count = {}
    for graph_op_info in graph_config:
        op_name = graph_op_info['op_name']
        op_index = graph_op_info['op_index']
        op_args = graph_op_info['op_args']
        op_kwargs = graph_op_info['op_kwargs']

        print(f'op_name {op_name}')
        input_ctx, output_ctx = op_index
        if not isinstance(input_ctx, tuple):
            input_ctx = (input_ctx,)
        if not isinstance(output_ctx, tuple):
            output_ctx = (output_ctx,)

        if op_name.startswith('deploy'):
            # 需要独立编译
            # 1.step 生成eagleeye算子封装
            op_name = op_name[7:]
            if op_name not in op_name_count:
                op_name_count[op_name] = 0
            op_unique_name = f'{op_name}_{op_name_count[op_name]}'
            op_name_count[op_name] += 1

            op_info = auto_generate_eagleeye_op(op_name, op_index, op_args, op_kwargs, output_folder)
            deploy_graph_info[op_unique_name] = op_info
        elif op_name.startswith('eagleeye'):
            # eagleeye核心算子 (op级别算子)
            op_name = op_name[9:]
            if op_name not in op_name_count:
                op_name_count[op_name] = 0
            op_unique_name = f'{op_name}_{op_name_count[op_name]}'
            op_name_count[op_name] += 1

            if op_name.endswith('_op'):
                op_name = op_name.capitalize()                
                op_name = op_name.replace('_op', 'Op')

            if op_name not in core_op_set:
                print(f'{op_name} not support')
                print(f'eagleeye core op set include {core_op_set.keys()}')
                return False

            # op_args， 包含scalar, numpy, list类型，转换成std::vector<float>类型
            deploy_graph_info[op_unique_name] = {
                'type': op_name,
                'input': input_ctx,
                'output': output_ctx,
                'args': convert_args_eagleeye_op_args(op_args, op_kwargs),
                'include': core_op_set[op_name]['include']
            }
        else:
            # eagleey核心算子 （op级别算子）
            if op_name not in op_name_count:
                op_name_count[op_name] = 0
            op_unique_name = f'{op_name}_{op_name_count[op_name]}'
            op_name_count[op_name] += 1

            if op_name.startswith('inference_onnx_op'):
                # 算子转换为平台预测引擎算子
                engine_op_name, template_args, op_args, include_path = \
                    convert_onnx_to_platform_engine(op_name, op_index, op_args, op_kwargs, output_folder, platform, abi, project_name=project_config['name'])

                deploy_graph_info[op_unique_name] = {
                    'type': engine_op_name,
                    'template': template_args,
                    'input': input_ctx,
                    'output': output_ctx,
                    'args': op_args,
                    'include': include_path
                }
                continue
            
            # 转换统一格式
            if op_name.endswith('_op'):
                op_name = op_name.capitalize()                     
                op_name = op_name.replace('_op', 'Op')

            # 检查是否在核心集中
            if op_name not in core_op_set:
                print(f'{op_name} not support')
                print(f'eagleeye core op set include {core_op_set.keys()}')
                return False

            # op_args， 包含scalar, numpy, list类型，转换成std::vector<float>类型
            deploy_graph_info[op_unique_name] = {
                'type': op_name,
                'input': input_ctx,
                'output': output_ctx,
                'args': convert_args_eagleeye_op_args(op_args, op_kwargs),
                'include': core_op_set[op_name]['include']
            }

    # 准备插件文件代码
    # 包括任务管线建立, nano算子图，任务输入信号设置，任务输出信号设置
    t = [v['include'] for v in deploy_graph_info.values()]
    include_list = ''
    for f in t:
        include_list += f'#include "{f}"\n'

    op_graph_code = ''
    deploy_output_data_name_inv_link = {}
    for deploy_op_name, deploy_op_info in deploy_graph_info.items():
        if 'template' in deploy_op_info:
            op_graph_code += f'dataflow::Node* {deploy_op_name} = op_graph->add("{deploy_op_name}", {deploy_op_info["type"]}{deploy_op_info["template"]}(), EagleeyeRuntime(EAGLEEYE_CPU));\n'
        else:
            op_graph_code += f'dataflow::Node* {deploy_op_name} = op_graph->add("{deploy_op_name}", {deploy_op_info["type"]}(), EagleeyeRuntime(EAGLEEYE_CPU));\n'

        deploy_op_args_tuple = deploy_op_info['args']
        if isinstance(deploy_op_args_tuple, dict):
            deploy_op_args_tuple = (deploy_op_args_tuple,)

        for deploy_op_args in deploy_op_args_tuple:
            arg_code = ''
            for deploy_arg_name, deploy_arg_list in deploy_op_args.items():
                if deploy_arg_name != 'c++_type':
                    if arg_code == '':
                        arg_code = '{"'+deploy_arg_name+'",{'+','.join([str(v) for v in deploy_arg_list])+'}}'
                    else:
                        arg_code += ',{"'+deploy_arg_name+'",{'+','.join([str(v) for v in deploy_arg_list])+'}}'

            if arg_code != '':
                args_init_code = deploy_op_args['c++_type']+'({'+arg_code+'})'
                op_graph_code += f'{deploy_op_name}->init({args_init_code});\n\n'

        print(deploy_op_info['output'])
        for data_i, data_name in enumerate(deploy_op_info['output']):
            deploy_output_data_name_inv_link[data_name] = (deploy_op_name, data_i)

    for deploy_op_name, deploy_op_info in deploy_graph_info.items():
        if deploy_op_info['input'] is not None:
            for input_data_i, input_data_name in enumerate(deploy_op_info['input']):
                if input_data_name is not None:
                    from_op_name, from_op_out_i = deploy_output_data_name_inv_link[input_data_name]
                    op_graph_code += f'op_graph->bind("{from_op_name}", {from_op_out_i}, "{deploy_op_name}", {input_data_i});\n'

    op_graph_code += 'op_graph->init(NULL);'

    eagleeye_plugin_code_content = \
        gen_code('./templates/plugin_code.cpp')(            
            project=project_config['name'],
            version=project_config.get('version', '1.0.0.0'),
            signature=project_config.get('signature', 'xxx'),
            include_list=include_list,
            in_port='{'+','.join([str(i) for i in range(len(project_config['input']))]) + '}',
            in_signal='{'+','.join(['"'+info[-1]+'"' for info in project_config['input']])+'}',
            out_port='{'+','.join([str(i) for i in range(len(project_config['output']))]) + '}',
            out_signal='{'+','.join(['"'+info[-1]+'"' for info in project_config['output']])+'}',
            graph_in_ops='{'+','.join(['"'+deploy_output_data_name_inv_link[info[0]][0]+'"' for info in project_config['input']])+'}',
            graph_out_ops='{'+','.join(['{"'+deploy_output_data_name_inv_link[info[0]][0]+'",'+str(deploy_output_data_name_inv_link[info[0]][1])+'}' for info in project_config['output']])+'}',
            op_graph=op_graph_code
        )

    with open(os.path.join(output_folder, f'{project_config["name"]}_plugin.cpp'), 'w') as fp:
        fp.write(eagleeye_plugin_code_content)

    # 准备额外依赖库（libc++_shared.so）
    if platform.lower() == 'android':
        ndk_path = os.environ['ANDROID_NDK_HOME']
        shutil.copy(os.path.join(ndk_path, "sources/cxx-stl/llvm-libc++/libs", abi, 'libc++_shared.so'), os.path.join(output_folder, '3rd', abi, 'libc++_shared.so'))

    # 更新CMakeLists.txt
    update_cmakelist(output_folder, project_config["name"], [s['src'] for s in deploy_graph_info.values() if 'src' in s])

    # 编译插件工程
    shell_code_content = gen_code('./templates/android_build.sh')(
        project=project_config['name'],
        ANDROID_NDK_HOME=os.environ['ANDROID_NDK_HOME']
    )
    with open(os.path.join(output_folder, 'android_build.sh'), 'w') as fp:
        fp.write(shell_code_content)

    shell_code_content = gen_code('./templates/linux_build.sh')(
        project=project_config['name']
    )
    with open(os.path.join(output_folder, 'linux_build.sh'), 'w') as fp:
        fp.write(shell_code_content)

    os.system(f'cd {output_folder} && bash {platform}_build.sh')


NANO_OP_REG = re.compile('^class\s*\w*\s*:')

def load_eagleeye_op_set(eagleeye_path):
    nano_op_path = os.path.join(eagleeye_path, 'include', "eagleeye", "engine", "nano", "op")
    op_set = {}
    for file_name in os.listdir(nano_op_path):
        if file_name.endswith('.h'):
            # 解析算子类名称
            for line in open(os.path.join(nano_op_path, file_name)):
                match = NANO_OP_REG.search(line)
                if match is not None:
                    op_name = match.group()[:-1]
                    op_name = re.split(r'\s+', op_name)[1]
                    op_set[op_name]={
                        'include': os.path.join('eagleeye', 'engine', 'nano', 'op', file_name)
                    }
    return op_set


def prepare_eagleeye_environment(system_platform, abi_platform):
    print('Check eagleeye environment')
    os.makedirs('.3rd/', exist_ok=True)
    if not os.path.exists(os.path.join('.3rd/', 'eagleeye')) or len(os.listdir(os.path.join('.3rd/', 'eagleeye'))) == 0:
        print('Download eagleeye git')
        os.system('cd .3rd/ && git clone https://github.com/jianzfb/eagleeye.git')

    p = subprocess.Popen("pip3 show eagleeye", shell=True, encoding="utf-8", stdout=subprocess.PIPE)
    if p.stdout.read() == '':
        print('Install eagleeye scafold')
        os.system('cd .3rd/eagleeye/scripts && pip3 install -r requirements.txt && python3 setup.py install')

    eagleeye_path = '.3rd/eagleeye/install'
    if not os.path.exists(eagleeye_path):
        print('Compile eagleeye core sdk')
        os.system(f'cd .3rd/eagleeye && bash {system_platform.lower()}_build.sh')
    eagleeye_path = os.path.abspath(eagleeye_path)
    return eagleeye_path


class DeployMixin:
    def build(self, platform='android/arm64-v8a', output_folder='./deploy', project_config=None):
        # android/arm64-v8a, linux/x86-64
        assert(platform in ['android/arm64-v8a', 'linux/x86-64'])
        system_platform, abi_platform = platform.split('/')

        # 准备eagleeye集成环境
        eagleeye_path = prepare_eagleeye_environment(system_platform, abi_platform)

        # 创建工程
        os.makedirs(output_folder, exist_ok=True)
        if not os.path.exists(os.path.join(output_folder, f'{project_config["name"]}_plugin')):
            if project_config.get('git', None) is not None and project_config['git'] != '':
                os.system(f'cd {output_folder} && git clone {project_config["git"]}')
            else:
                os.system(f'cd {output_folder} && eagleeye-cli project --project={project_config["name"]} --version={project_config.get("version", "1.0.0.0")} --signature=xxxxx --build_type=Release --abi={abi_platform.capitalize() if system_platform != "android" else abi_platform} --eagleeye={eagleeye_path}')
        output_folder = os.path.join(output_folder, f'{project_config["name"]}_plugin')

        # 编译
        package_build(
            output_folder, 
            eagleeye_path, 
            project_config=project_config, platform=system_platform, abi=abi_platform)

        return True




# -*- coding: UTF-8 -*-
# @Time    : 18-9-22
# @File    : batch.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

from antgo.dataflow.dataset.spider_dataset import SpiderDataset
from antgo.ant.base import *
from antgo.ant.base import _pick_idle_port
from antgo.task.task import *
from antgo.dataflow.common import *
from antgo.dataflow.basic import *
from antgo.dataflow.recorder import *
from antgo.crowdsource.batch_server import *
from antvis.client.httprpc import *
import json
import traceback
from antgo.dataflow.dataset.proxy_dataset import *
import threading


class AntBatch(AntBase):
    def __init__(self,
                 ant_context,
                 ant_name,
                 token,
                 data_factory,
                 ant_dump_dir,
                 ant_task_config,
                 dataset,
                 **kwargs):
        super(AntBatch, self).__init__(ant_name, ant_context, token)

        self.ant_data_source = data_factory
        self.ant_dump_dir = ant_dump_dir
        self.ant_context.ant = self
        self.unlabel = kwargs.get('unlabel', False)
        self.ant_task_config = ant_task_config
        self.restore_experiment = kwargs.get('restore_experiment', None)
        self.host_ip = self.context.params.system.get('ip', '127.0.0.1')
        self.host_port = self.context.params.system.get('port', -1)
        self.dataset = dataset
        self.rpc = None
        self._running_dataset = None
        self._running_task = None

    def ping_until_ok(self):
        while True:
            content = self.rpc.ping.get()
            if content['status'] != 'ERROR':
                break
            # 暂停5秒钟，再进行尝试
            time.sleep(5)

    @property
    def running_dataset(self):
        return self._running_dataset

    @property
    def running_task(self):
        return self._running_task

    def start(self):
        # 1.step 加载挑战任务
        running_ant_task = None
        if self.token is not None:
            # 1.1.step 从平台获取挑战任务配置信息
            response = mlogger.info.challenge.get(command=type(self).__name__)
            if response['status'] == 'ERROR':
                logger.error('Couldnt load challenge task.')
                self.token = None
            elif response['status'] == 'SUSPEND':
                # prohibit submit challenge task frequently
                # submit only one in one week
                logger.error('Prohibit submit challenge task frequently.')
                exit(-1)
            elif response['status'] == 'OK':
                content = response['content']

                if 'task' in content:
                    challenge_task = create_task_from_json(content)
                    if challenge_task is None:
                        logger.error('Couldnt load challenge task.')
                        exit(-1)
                    running_ant_task = challenge_task
            else:
                # unknow error
                logger.error('Unknow error.')
                exit(-1)

        if running_ant_task is None:
            # 1.2.step 加载自定义任务配置信息
            custom_task = create_task_from_xml(self.ant_task_config, self.context)
            if custom_task is None:
                logger.error('Couldnt load custom task.')
                exit(-1)
            running_ant_task = custom_task

        assert (running_ant_task is not None)
        self._running_task = running_ant_task

        # 2.step 获得实验ID
        experiment_uuid = self.context.experiment_uuid
        if self.restore_experiment is not None:
            experiment_uuid = self.restore_experiment

        # 3.step make experiment folder
        logger.info('Build experiment folder.')
        if not os.path.exists(os.path.join(self.ant_dump_dir, experiment_uuid)):
            os.makedirs(os.path.join(self.ant_dump_dir, experiment_uuid))
        experiment_dump_dir = os.path.join(self.ant_dump_dir, experiment_uuid)

        # 获得是否加载web server标记
        is_launch_web_server = True
        if self.host_port < 0:
            is_launch_web_server = False

        # 4.step 设置记录器
        # 设置记录器
        # 数据标签
        batch_params = getattr(self.context.params, 'predict', None)
        tags = []
        if batch_params is not None:
            tags = batch_params.get('tags', [])

        def _callback_func(dataset_stage, data):
            record_content = {
                'experiment_uuid': experiment_uuid,
                'dataset': {
                    dataset_stage: [{'data': single_d, 'tag': []} for single_d in data]
                },
                'tags': tags
            }
            for data_group in record_content['dataset'][dataset_stage]:
                for item in data_group['data']:
                    if item['type'] in ['IMAGE', 'VIDEO', 'FILE']:
                        item['data'] = 'static/data/%s/record/%s' % (dataset_stage, item['data'])

            # 添加当前进度
            record_content['waiting'] = \
                self.running_dataset.waiting_process_num() if self.running_dataset is not None else \
                    self.context.params.predict.get('size', 0)
            record_content['finished'] = \
                self.running_dataset.finish_process_num() if self.running_dataset is not None else 0
            self.rpc.predict.config.post(config_data=json.dumps(record_content))

            # 保存执行信息
            # 每执行完一次预测并记录，会将结果同时记录到json中
            if not os.path.exists(os.path.join(experiment_dump_dir, '%s.json' % experiment_uuid)):
                with open(os.path.join(experiment_dump_dir, '%s.json' % experiment_uuid), 'w') as fp:
                    json.dump(record_content, fp)
            else:
                history_running_info = {}
                with open(os.path.join(experiment_dump_dir, '%s.json' % experiment_uuid), 'r') as fp:
                    history_running_info = json.load(fp)
                history_running_info['dataset'][dataset_stage].extend(record_content['dataset'][dataset_stage])

                with open(os.path.join(experiment_dump_dir, '%s.json' % experiment_uuid), 'w') as fp:
                    json.dump(history_running_info, fp)

        output_dir = experiment_dump_dir
        if is_launch_web_server:
            output_dir = \
                os.path.join(experiment_dump_dir, 'predict', 'static', 'data', 'test')
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)

        self.context.recorder = \
            LocalRecorderNodeV2(lambda data: _callback_func('test', data) if is_launch_web_server else None)
        self.context.recorder.dump_dir = os.path.join(output_dir, 'record')
        if not os.path.exists(self.context.recorder.dump_dir):
            os.makedirs(self.context.recorder.dump_dir)

        # 4.step 配置web服务基本信息
        if is_launch_web_server:
            # 选择端口
            self.host_port = _pick_idle_port(self.host_port)

            # 配置调用
            self.rpc = HttpRpc("v1", "antgo/api", self.host_ip, self.host_port)

            # 准备素材资源
            static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
            batch_static_dir = os.path.join(experiment_dump_dir, 'predict', 'static')
            if os.path.exists(batch_static_dir):
                shutil.rmtree(batch_static_dir)

            shutil.copytree(os.path.join(static_folder, 'resource', 'app'), batch_static_dir)
            # 启动服务器
            white_users = self.context.params.predict.get('white_users', None)
            p = \
                multiprocessing.Process(
                    target=batch_server_start,
                    args=(
                        experiment_dump_dir,
                        self.host_port,
                        None,
                        white_users
                    ))
            p.start()

            # 等待web服务启动
            self.ping_until_ok()

        if self.context.is_interact_mode:
          logger.info('Running on interact mode.')
          return

        # 5.step 配置运行
        def _run_batch_process():
            # 是否来自已有实验
            if self.restore_experiment is not None and \
                os.path.exists(
                    os.path.join(experiment_dump_dir, '%s.json' % self.restore_experiment)):
                # 加载保存的执行信息
                running_info = {}
                with open(os.path.join(experiment_dump_dir, '%s.json' % self.restore_experiment), 'r') as fp:
                    running_info = json.load(fp)

                # 更新web页面
                if is_launch_web_server:
                    self.rpc.predict.config.post(config_data=json.dumps(running_info))
                return

            # 4.step load dataset
            logger.info('Load task dataset and split.')
            
            dataset_name = ''
            selected_dataset_stages = []
            if len(running_ant_task.dataset_name.split('/')) == 2:
                dataset_name, dataset_stage = \
                    running_ant_task.dataset_name.split('/')
                selected_dataset_stages.append(dataset_stage)
            else:
                dataset_name = running_ant_task.dataset_name
                selected_dataset_stages = ['test']

            if self.dataset is not None and self.dataset != '':
                logger.info('')
                if len(self.dataset.split('/')) == 2:
                    dataset_name, dataset_stage = self.dataset.split('/')
                    selected_dataset_stages.append(dataset_stage)
                else:
                    dataset_name = self.dataset
                    selected_dataset_stages = ['test']

            logger.info('Using dataset %s/%s'%(dataset_name, selected_dataset_stages[0]))

            # 5.step prepare ablation blocks
            logger.info('Prepare model ablation blocks.')
            ablation_blocks = getattr(self.ant_context.params, 'ablation', [])
            if ablation_blocks is None:
                ablation_blocks = []
            for b in ablation_blocks:
                self.ant_context.deactivate_block(b)

            # 6.step infer
            logger.info('Running inference process.')
            for dataset_stage in selected_dataset_stages:
                logger.info('Process dataset: %s.' % dataset_stage)
                # 获得数据集对象
                ant_test_dataset = None
                if dataset_name.startswith('spider'):
                    # 使用SpiderDataset
                    ant_test_dataset = \
                        SpiderDataset(self.command_queue,
                                        os.path.join(self.ant_data_source, dataset_name), None)
                else:
                    if self.context.register_at(dataset_stage) is not None:
                        ant_test_dataset = ProxyDataset(dataset_stage)
                        ant_test_dataset.register(**{dataset_stage:self.context.register_at(dataset_stage)})
                    else:
                        ant_test_dataset = \
                            running_ant_task.dataset(dataset_stage,
                                                        os.path.join(self.ant_data_source, dataset_name),
                                                        running_ant_task.dataset_params)

                # using unlabel
                if self.unlabel:
                    if 'candidate_file' in self.context.params.system['ext_params']:
                        logger.info('Using candidate file %s'%self.context.params.system['ext_params']['candidate_file'])
                        ant_test_dataset.candidate_file = self.context.params.system['ext_params']['candidate_file']
                    if 'unlabeled_list_file' in self.context.params.system['ext_params']:
                        logger.info('Using unlabeled list file %s'%self.context.params.system['ext_params']['unlabeled_list_file'])
                        ant_test_dataset.unlabeled_list_file = self.context.params.system['ext_params']['unlabeled_list_file']
                    ant_test_dataset = UnlabeledDataset(ant_test_dataset)

                # 重新设置输出目录
                output_dir = experiment_dump_dir
                if is_launch_web_server:
                    output_dir = \
                        os.path.join(experiment_dump_dir, 'predict', 'static', 'data', dataset_stage)
                    if not os.path.exists(output_dir):
                        os.makedirs(output_dir)

                # 重新设置记录器
                self.context.recorder = \
                    LocalRecorderNodeV2(lambda data: _callback_func(dataset_stage, data) if is_launch_web_server else None)

                # 调用预测的回调函数
                self._running_dataset = ant_test_dataset
                with safe_recorder_manager(self.context.recorder):
                    # 完成推断过程
                    try:
                        self.context.call_infer_process(ant_test_dataset, dump_dir=output_dir)
                    except Exception as e:
                        if type(e.__cause__) != StopIteration:
                            print(e)
                            traceback.print_exc()

                self.context.recorder.close()
            return

        # 5.step 启动bach server
        _run_batch_process()


exp=dict(
    launch_script='python3 test_m.py',
    args='',
    schedule=[
        {
            '--checkpoint': '/opt/tiger/handdetJ/checkpoint/greedy_ac_1/handdetJ/epoch_50.pth',
        }
    ]
)


#encoding=utf-8
from __future__ import unicode_literals
from __future__ import division

import numpy as np
import unittest

def mape(actual_s,predicated_s):
    '''
    MAPE(median absolute percentage error) could be subjective to outliers(for Regresion Task)
    RMSE has some problems, which is sensitive to large outliers. MAPE maybe a good select.
    Parameters
    ----------
    actual_s: list
            ground truth value
    predicated_s: list
            predicated value

    Returns
    -------
    error
    '''
    if type(actual_s) == list:
        actual_s = np.array(actual_s, dtype=np.float32)
    if type(predicated_s) == list:
        predicated_s = np.array(predicated_s, dtype=np.float32)

    if len(predicated_s.shape) > 1 and predicated_s.shape[1] > 1:
        predicated_s = predicated_s[:,0]
    #reshape
    predicated_s = np.reshape(predicated_s,[predicated_s.size])

    diff = actual_s - predicated_s
    actual_s_eps = actual_s + 0.000000001   #avoid 0
    v = np.median(np.abs(diff / actual_s_eps))

    return v


def almost_correct(actual_s,predicated_s,X):
    '''
    It compute the percent of estimates that differ from the true value by
    no more than X%
    Parameters
    ----------
    actual_s: list
            ground truth value
    predicated_s: list
            predicated value
    X:  double
            percentage

    Returns
    -------
    score: double
            the percent of estimates within X% of the true values

    '''
    if type(actual_s) == list:
        actual_s = np.array(actual_s, dtype=np.float32)
    if type(predicated_s) == list:
        predicated_s = np.array(predicated_s, dtype=np.float32)

    #absorb some error
    if len(predicated_s.shape) > 1 and predicated_s.shape[1] > 1:
        predicated_s = predicated_s[:,0]
    #reshape
    predicated_s = np.reshape(predicated_s,[predicated_s.size])

    diff = actual_s - predicated_s
    actual_s_eps = actual_s + 0.000000001  # avoid 0

    diff = np.abs(diff/actual_s_eps)
    num = actual_s.shape[0]
    return float(len(np.where(diff <= X)[0])) / float(num)

class TestMAPEandAC(unittest.TestCase):
    def test(self):
        actual_s = [0.1,0.2,0.1,0.8,0.3]
        predicated_s = [0.12,0.1,0.11,0.74,0.04]
        self.assertAlmostEqual(mape(actual_s,predicated_s),0.2)
        self.assertAlmostEqual(almost_correct(actual_s, predicated_s, 0.1), 0.4)

if __name__ == "__main__":
    unittest.main()


# -----------------------------------------------------
# Copyright (c) Shanghai Jiao Tong University. All rights reserved.
# Written by Jiefeng Li (jeff.lee.sjtu@gmail.com)
# -----------------------------------------------------

import torch.nn as nn
import torch.nn.functional as F


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
        dcn=None,
    ):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.relu2 = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu1(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu2(out)

        return out


class InverseBottleneckDwcK5(nn.Module):
    expansion = 1
    exp = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
        dcn=None,
    ):
        super(InverseBottleneckDwcK5, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("InverseBottleneck only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in InverseBottleneck")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = nn.Conv2d(inplanes, planes * self.exp, 1, 1, 0)
        self.bn1 = norm_layer(planes * self.exp)
        self.relu1 = nn.ReLU(inplace=True)
        if dilation == 0:
            self.conv2 = nn.Sequential(
                nn.Conv2d(
                    planes * self.exp, planes * self.exp, kernel_size=5, groups=planes * self.exp, padding=1, bias=False
                ),
                norm_layer(planes * self.exp),
                nn.Conv2d(
                    planes * self.exp, planes * self.exp, kernel_size=5, groups=planes * self.exp, padding=1, bias=False
                ),
                norm_layer(planes * self.exp),
            )
        else:
            self.conv2 = conv3x3(planes * self.exp, planes * self.exp, stride, planes * self.exp, dilation=dilation)
        self.bn2 = norm_layer(planes * self.exp)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv3 = nn.Conv2d(planes * self.exp, planes, 1, 1, 0)
        self.bn3 = norm_layer(planes)
        self.relu3 = nn.ReLU(inplace=True)

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu1(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu2(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu3(out)

        return out


class InverseBottleneck(nn.Module):
    expansion = 1
    exp = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
        dcn=None,
    ):
        super(InverseBottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("InverseBottleneck only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in InverseBottleneck")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = nn.Conv2d(inplanes, planes * self.exp, 1, 1, 0)
        self.bn1 = norm_layer(planes * self.exp)
        self.relu1 = nn.ReLU(inplace=True)
        if dilation == 0:
            self.conv2 = nn.Sequential(
                nn.Conv2d(planes * self.exp, planes * self.exp, kernel_size=3, groups=planes * self.exp, bias=False),
                norm_layer(planes * self.exp),
                nn.Conv2d(planes * self.exp, planes * self.exp, kernel_size=3, groups=planes * self.exp, bias=False),
                norm_layer(planes * self.exp),
            )
        else:
            self.conv2 = conv3x3(planes * self.exp, planes * self.exp, stride, planes * self.exp, dilation=dilation)
        self.bn2 = norm_layer(planes * self.exp)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv3 = nn.Conv2d(planes * self.exp, planes, 1, 1, 0)
        self.bn3 = norm_layer(planes)
        self.relu3 = nn.ReLU(inplace=True)

        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu1(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu2(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu3(out)

        return out


class Bottleneck(nn.Module):
    # expansion = 4
    expansion = 2

    def __init__(self, inplanes, planes, stride=1, downsample=None, norm_layer=nn.BatchNorm2d, dcn=None):
        super(Bottleneck, self).__init__()
        self.dcn = dcn
        self.with_dcn = dcn is not None

        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = norm_layer(planes, momentum=0.1)
        if self.with_dcn:
            fallback_on_stride = dcn.get("FALLBACK_ON_STRIDE", False)
            self.with_modulated_dcn = dcn.get("MODULATED", False)
        if not self.with_dcn or fallback_on_stride:
            self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        else:
            self.deformable_groups = dcn.get("DEFORM_GROUP", 1)
            if not self.with_modulated_dcn:
                conv_op = DeformConv
                offset_channels = 18
            else:
                conv_op = ModulatedDeformConv
                offset_channels = 27

            self.conv2_offset = nn.Conv2d(
                planes, self.deformable_groups * offset_channels, kernel_size=3, stride=stride, padding=1
            )
            self.conv2 = conv_op(
                planes,
                planes,
                kernel_size=3,
                stride=stride,
                padding=1,
                deformable_groups=self.deformable_groups,
                bias=False,
            )

        self.bn2 = norm_layer(planes, momentum=0.1)
        self.conv3 = nn.Conv2d(planes, planes * 2, kernel_size=1, bias=False)  # 4->2
        self.bn3 = norm_layer(planes * 2, momentum=0.1)  # 4->2
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = F.relu(self.bn1(self.conv1(x)), inplace=True)
        if not self.with_dcn:
            out = F.relu(self.bn2(self.conv2(out)), inplace=True)
        elif self.with_modulated_dcn:
            offset_mask = self.conv2_offset(out)
            offset = offset_mask[:, : 18 * self.deformable_groups, :, :]
            mask = offset_mask[:, -9 * self.deformable_groups :, :, :]
            mask = mask.sigmoid()
            out = F.relu(self.bn2(self.conv2(out, offset, mask)))
        else:
            offset = self.conv2_offset(out)
            out = F.relu(self.bn2(self.conv2(out, offset)), inplace=True)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = F.relu(out)

        return out


class KetNetFSmall(nn.Module):
    """KetNetF"""

    def __init__(
        self,
        architecture,
        norm_layer=nn.BatchNorm2d,
        dcn=None,
        stage_with_dcn=(False, False, False, False, False, False, False, False),
    ):
        super(KetNetFSmall, self).__init__()
        self._norm_layer = norm_layer
        assert architecture in ["resnet18", "resnet34", "resnet50", "resnet101", "resnet152"]
        layers = {
            "resnet34": [1, 1, 1, 2, 2, 1, 4, 4],
        }
        self.inplanes = 32
        if architecture == "resnet18" or architecture == "resnet34":
            self.block = InverseBottleneck
        else:
            self.block = InverseBottleneck
        self.layers = layers[architecture]
        stage_dcn = [dcn if with_dcn else None for with_dcn in stage_with_dcn]

        self.conv1 = nn.Conv2d(1, 32, kernel_size=5, stride=2, padding=2, bias=False)
        self.bn1 = norm_layer(32, eps=1e-5, momentum=0.1, affine=True)
        self.relu1 = nn.ReLU(inplace=True)

        self.block.exp = 3
        self.layer1 = self.make_layer(self.block, 32, self.layers[0], stride=1, dcn=stage_dcn[0])
        self.layer2 = self.make_layer(self.block, 32, self.layers[1], stride=2, dcn=stage_dcn[1])
        self.block.exp = 6
        self.layer3 = self.make_layer(self.block, 32, self.layers[2], stride=1, dcn=stage_dcn[2])
        self.layer4 = self.make_layer(self.block, 64, self.layers[3], stride=2, dcn=stage_dcn[3])
        self.block.exp = 3
        self.layer5 = self.make_layer(self.block, 96, self.layers[4], stride=1, dcn=stage_dcn[4])
        self.layer6 = self.make_layer(self.block, 128, self.layers[5], stride=2, dcn=stage_dcn[5], keep_4=True)
        self.block.exp = 4
        self.layer7 = self.make_layer(self.block, 128, self.layers[6], stride=1, dcn=stage_dcn[6])
        self.layer8 = self.make_layer(self.block, 160, self.layers[7], stride=2, dcn=stage_dcn[7])

    def forward(self, x):
        x = self.relu1(self.bn1(self.conv1(x)))  # 32 * h/4 * w/4
        x = self.layer1(x)
        x32 = self.layer2(x)  # 32 * h/4 * w/4
        x32 = self.layer3(x32)  # 64 * h/4 * w/4
        x16 = self.layer4(x32)  # 64 * h/8 * w/8
        x16 = self.layer5(x16)  # 96 * h/8 * w/8
        x8 = self.layer6(x16)  # 128 * h/16 * w/16
        x8 = self.layer7(x8)  # 128 * h/16 * w/16
        x4 = self.layer8(x8)  # 160 * h/32 * w/32
        return [x32, x16, x8, x4]

    def stages(self):
        return [self.layer1, self.layer2, self.layer3, self.layer4]

    def make_layer(self, block, planes, blocks, stride=1, dcn=None, keep_4=False):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if keep_4:
                downsample = nn.Sequential(
                    nn.Conv2d(
                        self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, padding=2, bias=False
                    ),
                    self._norm_layer(planes * block.expansion),
                )
            else:
                downsample = nn.Sequential(
                    nn.Conv2d(self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                    self._norm_layer(planes * block.expansion),
                )

        layers = []
        if keep_4:
            layers.append(
                block(self.inplanes, planes, stride, downsample, norm_layer=self._norm_layer, dcn=dcn, dilation=0)
            )
        else:
            layers.append(block(self.inplanes, planes, stride, downsample, norm_layer=self._norm_layer, dcn=dcn))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, norm_layer=self._norm_layer, dcn=dcn))

        return nn.Sequential(*layers)


class KetNetFBig(nn.Module):
    """KetNetF"""

    def __init__(
        self,
        architecture,
        norm_layer=nn.BatchNorm2d,
        dcn=None,
        stage_with_dcn=(False, False, False, False, False, False, False, False),
    ):
        super(KetNetFBig, self).__init__()
        self._norm_layer = norm_layer
        assert architecture in ["resnet18", "resnet34", "resnet50", "resnet101", "resnet152"]
        layers = {
            "resnet34": [1, 1, 1, 2, 2, 1, 4, 4],
        }
        self.inplanes = 32
        if architecture == "resnet18" or architecture == "resnet34":
            self.block = InverseBottleneck
        else:
            self.block = InverseBottleneck
        self.layers = layers[architecture]
        stage_dcn = [dcn if with_dcn else None for with_dcn in stage_with_dcn]

        self.conv1 = nn.Conv2d(1, 32, kernel_size=5, stride=2, padding=2, bias=False)
        self.bn1 = norm_layer(32, eps=1e-5, momentum=0.1, affine=True)
        self.relu1 = nn.ReLU(inplace=True)

        self.block.exp = 3
        self.layer1 = self.make_layer(self.block, 32, self.layers[0], stride=1, dcn=stage_dcn[0])
        self.layer2 = self.make_layer(self.block, 32, self.layers[1], stride=2, dcn=stage_dcn[1])
        self.block.exp = 6
        self.layer3 = self.make_layer(self.block, 32, self.layers[2], stride=1, dcn=stage_dcn[2])
        self.layer4 = self.make_layer(self.block, 64, self.layers[3], stride=2, dcn=stage_dcn[3])
        self.block.exp = 4
        self.layer5 = self.make_layer(self.block, 96, self.layers[4], stride=1, dcn=stage_dcn[4])
        self.layer6 = self.make_layer(self.block, 128, self.layers[5], stride=2, dcn=stage_dcn[5], keep_4=True)
        self.block.exp = 6
        self.layer7 = self.make_layer(self.block, 128, self.layers[6], stride=1, dcn=stage_dcn[6])
        self.layer8 = self.make_layer(self.block, 160, self.layers[7], stride=2, dcn=stage_dcn[7])

    def forward(self, x):
        x = self.relu1(self.bn1(self.conv1(x)))  # 32 * h/4 * w/4
        x = self.layer1(x)
        x32 = self.layer2(x)  # 32 * h/4 * w/4
        x32 = self.layer3(x32)  # 64 * h/4 * w/4
        x16 = self.layer4(x32)  # 64 * h/8 * w/8
        x16 = self.layer5(x16)  # 96 * h/8 * w/8
        x8 = self.layer6(x16)  # 128 * h/16 * w/16
        x8 = self.layer7(x8)  # 128 * h/16 * w/16
        x4 = self.layer8(x8)  # 160 * h/32 * w/32
        return [x32, x16, x8, x4]

    def stages(self):
        return [self.layer1, self.layer2, self.layer3, self.layer4]

    def make_layer(self, block, planes, blocks, stride=1, dcn=None, keep_4=False):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if keep_4:
                downsample = nn.Sequential(
                    nn.Conv2d(
                        self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, padding=2, bias=False
                    ),
                    self._norm_layer(planes * block.expansion),
                )
            else:
                downsample = nn.Sequential(
                    nn.Conv2d(self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                    self._norm_layer(planes * block.expansion),
                )

        layers = []
        if keep_4:
            layers.append(
                block(self.inplanes, planes, stride, downsample, norm_layer=self._norm_layer, dcn=dcn, dilation=0)
            )
        else:
            layers.append(block(self.inplanes, planes, stride, downsample, norm_layer=self._norm_layer, dcn=dcn))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, norm_layer=self._norm_layer, dcn=dcn))

        return nn.Sequential(*layers)


class KetNetFBigDwcK5(nn.Module):
    """KetNetF"""

    def __init__(
        self,
        architecture,
        norm_layer=nn.BatchNorm2d,
        dcn=None,
        stage_with_dcn=(False, False, False, False, False, False, False, False),
    ):
        super(KetNetFBigDwcK5, self).__init__()
        self._norm_layer = norm_layer
        assert architecture in ["resnet18", "resnet34", "resnet50", "resnet101", "resnet152"]
        layers = {
            "resnet34": [1, 1, 1, 2, 2, 1, 4, 4],
        }
        self.inplanes = 32
        if architecture == "resnet18" or architecture == "resnet34":
            self.block = InverseBottleneckDwcK5
        else:
            self.block = InverseBottleneckDwcK5
        self.layers = layers[architecture]
        stage_dcn = [dcn if with_dcn else None for with_dcn in stage_with_dcn]

        self.conv1 = nn.Conv2d(1, 32, kernel_size=5, stride=2, padding=2, bias=False)
        self.bn1 = norm_layer(32, eps=1e-5, momentum=0.1, affine=True)
        self.relu1 = nn.ReLU(inplace=True)

        self.block.exp = 3
        self.layer1 = self.make_layer(self.block, 32, self.layers[0], stride=1, dcn=stage_dcn[0])
        self.layer2 = self.make_layer(self.block, 32, self.layers[1], stride=2, dcn=stage_dcn[1])
        self.block.exp = 6
        self.layer3 = self.make_layer(self.block, 32, self.layers[2], stride=1, dcn=stage_dcn[2])
        self.layer4 = self.make_layer(self.block, 64, self.layers[3], stride=2, dcn=stage_dcn[3])
        self.block.exp = 4
        self.layer5 = self.make_layer(self.block, 96, self.layers[4], stride=1, dcn=stage_dcn[4])
        self.layer6 = self.make_layer(self.block, 128, self.layers[5], stride=2, dcn=stage_dcn[5], keep_4=True)
        self.block.exp = 6
        self.layer7 = self.make_layer(self.block, 128, self.layers[6], stride=1, dcn=stage_dcn[6])
        self.layer8 = self.make_layer(self.block, 160, self.layers[7], stride=2, dcn=stage_dcn[7])

    def forward(self, x):
        x = self.relu1(self.bn1(self.conv1(x)))  # 32 * h/4 * w/4
        x = self.layer1(x)
        x32 = self.layer2(x)  # 32 * h/4 * w/4
        x32 = self.layer3(x32)  # 64 * h/4 * w/4
        x16 = self.layer4(x32)  # 64 * h/8 * w/8
        x16 = self.layer5(x16)  # 96 * h/8 * w/8
        x8 = self.layer6(x16)  # 128 * h/16 * w/16
        x8 = self.layer7(x8)  # 128 * h/16 * w/16
        x4 = self.layer8(x8)  # 160 * h/32 * w/32
        return [x32, x16, x8, x4]

    def stages(self):
        return [self.layer1, self.layer2, self.layer3, self.layer4]

    def make_layer(self, block, planes, blocks, stride=1, dcn=None, keep_4=False):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if keep_4:
                downsample = nn.Sequential(
                    nn.Conv2d(
                        self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, padding=2, bias=False
                    ),
                    self._norm_layer(planes * block.expansion),
                )
            else:
                downsample = nn.Sequential(
                    nn.Conv2d(self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                    self._norm_layer(planes * block.expansion),
                )

        layers = []
        if keep_4:
            layers.append(
                block(self.inplanes, planes, stride, downsample, norm_layer=self._norm_layer, dcn=dcn, dilation=0)
            )
        else:
            layers.append(block(self.inplanes, planes, stride, downsample, norm_layer=self._norm_layer, dcn=dcn))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, norm_layer=self._norm_layer, dcn=dcn))

        return nn.Sequential(*layers)



import torch.nn as nn
import torch

import copy

from antgo.framework.helper.models.nas.utils import download_url


# Helper for constructing the one-hot vectors.
def construct_maps(keys):
    d = dict()
    keys = list(set(keys))
    for k in keys:
        if k not in d:
            d[k] = len(list(d.keys()))
    return d


ks_map = construct_maps(keys=(3, 5, 7))
ex_map = construct_maps(keys=(3, 4, 6))
dp_map = construct_maps(keys=(2, 3, 4))


class AccuracyPredictor:
    def __init__(self, pretrained=True, device="cuda:0"):
        self.device = device

        self.model = nn.Sequential(
            nn.Linear(128, 400),
            nn.ReLU(),
            nn.Linear(400, 400),
            nn.ReLU(),
            nn.Linear(400, 400),
            nn.ReLU(),
            nn.Linear(400, 1),
        )
        if pretrained:
            # load pretrained model
            fname = download_url(
                "https://hanlab.mit.edu/files/OnceForAll/tutorial/acc_predictor.pth"
            )
            self.model.load_state_dict(
                torch.load(fname, map_location=torch.device("cpu"))
            )
        self.model = self.model.to(self.device)

    # TODO: merge it with serialization utils.
    @torch.no_grad()
    def predict_accuracy(self, population):
        all_feats = []
        for sample in population:
            ks_list = copy.deepcopy(sample["ks"])
            ex_list = copy.deepcopy(sample["e"])
            d_list = copy.deepcopy(sample["d"])
            r = copy.deepcopy(sample["r"])[0]
            feats = (
                AccuracyPredictor.spec2feats(ks_list, ex_list, d_list, r)
                .reshape(1, -1)
                .to(self.device)
            )
            all_feats.append(feats)
        all_feats = torch.cat(all_feats, 0)
        pred = self.model(all_feats).cpu()
        return pred

    @staticmethod
    def spec2feats(ks_list, ex_list, d_list, r):
        # This function converts a network config to a feature vector (128-D).
        start = 0
        end = 4
        for d in d_list:
            for j in range(start + d, end):
                ks_list[j] = 0
                ex_list[j] = 0
            start += 4
            end += 4

        # convert to onehot
        ks_onehot = [0 for _ in range(60)]
        ex_onehot = [0 for _ in range(60)]
        r_onehot = [0 for _ in range(8)]

        for i in range(20):
            start = i * 3
            if ks_list[i] != 0:
                ks_onehot[start + ks_map[ks_list[i]]] = 1
            if ex_list[i] != 0:
                ex_onehot[start + ex_map[ex_list[i]]] = 1

        r_onehot[(r - 112) // 16] = 1
        return torch.Tensor(ks_onehot + ex_onehot + r_onehot)



###########################################################
# 定制化绑定本地后端配置，如文件HDFS系统和KV系统
###########################################################

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.ant import environment
import os
import subprocess


# step1: 配置hdfs后端存储
class YourHDFS(environment.HDFSClient):
    def get(self, remote_path, local_path):
        # 下载远程文件到本地
        # remote_path: 远程hdfs地址
        # local_path: 本地地址
        os.system(f'hdfs dfs -get {remote_path} {local_path}')
        return True

    def rm(self, remote_path, is_folder=False):
        # 删除远程文件
        # remote_path: 远程hdfs地址
        return True

    def ls(self, remote_path):
        # 获得远程地址下的文件列表
        result = subprocess.Popen(f'hdfs dfs -ls {remote_path}', shell=True, stdout=subprocess.PIPE).stdout
        file_list = []
        for line_info in result.read().decode('utf-8').split('\n'):
            if 'hdfs' in line_info:
                a = line_info.split('hdfs://')[-1].strip()
                file_list.append(f'hdfs://{a}')
            
        return file_list

    def mkdir(self, remote_path, p=False):
        # 在远程地址下创建目录
        if p:
            os.system(f'hdfs dfs -mkdir -p {remote_path}')
        else:
            os.system(f'hdfs dfs -mkdir {remote_path}')
        return True

    def put(self, remote_path, local_path, is_exist=False):
        # 上传本地文件到远程地址
        if is_exist:
            os.system(f'hdfs dfs -put {local_path} {remote_path}')
        else:
            os.system(f'hdfs dfs -put -f {local_path} {remote_path}')
        return True

    def mv(self, remote_src_path, remote_target_path):
        # 移动远程src地址到tgt地址
        return True

    def exists(self, remote_path):
        # 检查远程地址是否存在
        result = subprocess.Popen(f'hdfs dfs -ls {remote_path}', shell=True, stdout=subprocess.PIPE).stdout
        if result.read() == b'':
            return False
        else:
            return True

environment.hdfs_client = YourHDFS()


# step2: 配置KV Reader & Writer
class YourKVReader(object):
    def __init__(self, *args, **kwargs):
        pass
    
    def read_many(self, keys):
        # keys: ["", ""]
        # 基于传入的keys读取
        return []


class YourKVWriter(object):
    def __init__(self, *args, **kwargs):
        pass

    def write_many(self, keys, values):
        # keys: ["", ""], values: [b'', b'']
        # 将keys,values写入
        pass

    def flush(self):
        # 刷新写缓存
        pass

# environment.KVReader = YourKVReader
# environment.KVWriter = YourKVWriter


from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import argparse
import importlib
import os
import sys
from pprint import pprint
parser = argparse.ArgumentParser(description=f'ANTGO')

def DEFINE_int(name, default, var_help):
    global parser
    parser.add_argument(
        f'--{name}',
        type=int,
        default=default,
        help=var_help
    )

def DEFINE_float(name, default, var_help):
    global parser
    parser.add_argument(
        f'--{name}',
        type=float,
        default=default,
        help=var_help
    )

def DEFINE_indicator(name, default, var_help):
    global parser
    parser.add_argument(
        f'--{name}',
        action='store_true' if default else 'store_false',
        help=var_help
    )   

def DEFINE_string(name, default, var_help):
    global parser
    parser.add_argument(
        f'--{name}',
        type=str,
        default=default,
        help=var_help
    )   

def DEFINE_choices(name, default, choices, var_help):
    global parser
    parser.add_argument(
        f'--{name}',
        choices=choices,
        default=default,
        help=var_help
    )    


def DEFINE_nn_args():
    global parser
    group_gpus = parser.add_mutually_exclusive_group()
    parser.add_argument('--config', type=str, default="config.py", help='train config file path')
    parser.add_argument('--extra-config', type=str, default="", help='train extra config file path')
    parser.add_argument(
        '--resume-from', default=None, help='the checkpoint file to resume from')

    parser.add_argument(
        '--auto-resume',
        action='store_true',
        help='resume from the latest checkpoint automatically')
    parser.add_argument(
        '--no-validate',
        action='store_true',
        help='whether not to evaluate the checkpoint during training')
    group_gpus.add_argument(
        '--gpu-id',
        type=str,
        default='0',
        help='id of gpu to use '
        '(only applicable to non-distributed training)')
    parser.add_argument('--seed', type=int, default=None, help='random seed')
    parser.add_argument(
        '--diff-seed',
        action='store_true',
        help='Whether or not set different seeds for different ranks')
    parser.add_argument(
        '--deterministic',
        action='store_true',
        help='whether to set deterministic options for CUDNN backend.')        
    parser.add_argument('--local_rank', type=int, default=0)
    parser.add_argument(
        '--auto-scale-lr',
        action='store_true',
        help='enable automatically scaling LR.')
    parser.add_argument(
        '--distributed',
        action='store_true',
        help='is distributed.')        
    parser.add_argument(
        '--fuse-conv-bn',
        action='store_true',
        help='fuse conv and bn'
    )
    parser.add_argument(
        '--work-dir',
        type=str,
        default='',
        help='work dir'
    )    
    parser.add_argument(
        '--checkpoint',
        type=str,
        default=None,
        help='checkpoint'
    ) 
    parser.add_argument(
        '--process',
        choices=['train', 'test', 'export', 'activelearning'],
        default='train',
        help='train or test process'
    )
    parser.add_argument(
        '--max-epochs',
        type=int,
        default=-1,
        help='train max epochs'
    )    
    parser.add_argument(
        '--root',
        type=str,
        default='',
        help='root path'        
    )
    parser.add_argument(
        '--exp',
        type=str,
        default='',
        help='experiment name'
    )
    parser.add_argument(
        '--find-unused-parameters',
        action='store_true',
        help='only work on multi gpu training process'
    )
    parser.add_argument('--ext-module', type=str, default='', help='introduce ext module py file')


def parse_args():
    global parser
    args = parser.parse_args()

    if 'LOCAL_RANK' not in os.environ:
        if getattr(args, 'local_rank', None):
           os.environ['LOCAL_RANK'] = str(args.local_rank)
    return args

def print_args(s):
    s_dict = {}
    for k,v in s._get_kwargs():
        s_dict[k] = v
    pprint(f'exp {s.exp} config')
    pprint(s_dict)



# -*- coding: UTF-8 -*-
# @Time    : 2019/1/23 3:13 PM
# @File    : kcenter_greedy.py
# @Author  : jian<jian@mltalker.com>

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

"""Returns points that minimizes the maximum distance of any point to a center.

Implements the k-Center-Greedy method in
Ozan Sener and Silvio Savarese.  A Geometric Approach to Active Learning for
Convolutional Neural Networks. https://arxiv.org/abs/1708.00489 2017

Distance metric defaults to l2 distance.  Features used to calculate distance
are either raw features or if a model has transform method then uses the output
of model.transform(X).

Can be extended to a robust k centers algorithm that ignores a certain number of
outlier datapoints.  Resulting centers are solution to multiple integer program.
"""

import numpy as np
from sklearn.metrics import pairwise_distances
from .sampling_def import SamplingMethod



class kCenterGreedy(SamplingMethod):
  def __init__(self, X, metric='euclidean'):
    self.name = 'kcenter'
    self.features = self.flatten_X(X)
    self.metric = metric
    self.min_distances = None
    self.n_obs = self.features.shape[0]

  def update_distances(self, cluster_centers, already_selected, only_new=True, reset_dist=False):
    """Update min distances given cluster centers.

    Args:
      cluster_centers: indices of cluster centers
      only_new: only calculate distance for newly selected points and update
        min_distances.
      rest_dist: whether to reset min_distances.
    """

    if reset_dist:
      self.min_distances = None
    if only_new:
      cluster_centers = [d for d in cluster_centers
                         if d not in already_selected]
    if cluster_centers:
      # Update min_distances for all examples given new cluster center.
      x = self.features[cluster_centers]
      dist = pairwise_distances(self.features, x, metric=self.metric)

      if self.min_distances is None:
        self.min_distances = np.min(dist, axis=1).reshape(-1,1)
      else:
        self.min_distances = np.minimum(self.min_distances, dist)

  def select_batch_(self, already_selected, N, **kwargs):
    """
    Diversity promoting active learning method that greedily forms a batch
    to minimize the maximum distance to a cluster center among all unlabeled
    datapoints.

    Args:
      already_selected: index of datapoints already selected
      N: batch size

    Returns:
      indices of points selected to minimize distance to cluster centers
    """
    # init
    self.update_distances(None, None, only_new=False, reset_dist=True)

    new_batch = []
    for _ in range(N):
      if self.min_distances is None:
        # Initialize centers with a randomly selected datapoint
        ind = np.random.choice(np.arange(self.n_obs))
      else:
        ind = np.argmax(self.min_distances)
      # New examples should not be in already selected since those points
      # should have min_distance of zero to a cluster center.
      assert ind not in already_selected

      self.update_distances([ind], already_selected, only_new=True, reset_dist=False)
      new_batch.append(ind)

    print('Maximum distance from cluster centers is %0.2f'% max(self.min_distances))
    return new_batch



from .hrnet import HRNet
from .resnet import ResNet, ResNetV1d
from .mobilenet_v2 import MobileNetV2
from .resnext import ResNeXt

__all__ = [
    'HRNet','ResNet', 'ResNetV1d', 'MobileNetV2', 'ResNeXt'
]


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 22:44
# @File    : option.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from typing import Callable, Generic
from typing import Any, TypeVar

A = TypeVar('A', covariant=True)
B = TypeVar('B')
T = TypeVar('T')


class _Reason:
    """
    reason for `Empty` value
    """

    def __init__(self, x: Any, e: Exception) -> None:
        self._value = x
        self._exception = e

    @property
    def value(self):
        return self._value

    @property
    def exception(self):
        return self._exception


class Option(Generic[A]):
    """
    Functional-style error handling.

    Option[A] = Some(A) or Empty[A]
    1. Some(A): just a container for result;
    2. Empty[A]: result is empty, because of input error or computation error;

    Examples:

    >>> a = Some(10)
    >>> a.map(lambda x: x/2.0)
    Some(5.0)

    >>> a.map(lambda x: x/0)
    Empty()

    >>> b = Empty()
    >>> b.map(lambda x: x/2.0)
    Empty()
    """

    @staticmethod
    def of(x: T):
        """Return a boxed value

        Args:
            x (T): input value

        Returns:
            Some(T): boxed value
        """
        return Some(x)

    @staticmethod
    def empty():
        """Return an empty value

        Returns:
            Empty: empty value
        """
        return Empty()

    def flat_map(self, f: Callable[[A], 'Option[B]']) -> 'Option[B]':
        """Apply boxed version of callable

        Args:
            f (Callable[[A], Option[B]]): boxed version of callable

        Returns:
            Option[B]: boxed value

        Examples:

        >>> Option.of(1).flat_map(lambda x: x+1)
        2

        >>> Option.empty().flat_map(lambda x: x+1)
        Empty()
        """
        if isinstance(self, Some):
            return f(self._value)
        else:
            return self

    def map(self, f: Callable[[A], 'B']) -> 'Option[B]':
        """Apply function to value

        Args:
            f (Callable[[A], B]): unboxed function

        Returns:
            Option[B]: boxed return value
        """

        def wrapper(x):
            try:
                return Some(f(x))
            except Exception as e:  # pylint: disable=broad-except
                return Empty(x, e)

        return self.flat_map(wrapper)

    def is_empty(self):
        """Return True if the value is empty.
        """
        return isinstance(self, Empty)

    def is_some(self):
        """Return True if the value is boxed value.
        """
        return isinstance(self, Some)

    def get_or_else(self, default):
        """Return unboxed value, or default is the value is empty.

        Examples:

        >>> Option.of(0).get_or_else(1)
        0

        >>> Option.empty().get_or_else(1)
        1
        """
        if self.is_some():
            return self.get()
        return default


class Some(Option[A]):
    """
    `Some` value for `Option`
    """

    def __init__(self, x: A) -> None:
        self._value = x

    def __repr__(self) -> str:
        return 'Some({})'.format(self._value)

    def flat_map(self, f: Callable[[A], 'Option[B]']) -> 'Option[B]':
        return f(self._value)

    def get(self):
        """Return unboxed value
        """
        return self._value


class Empty(Option[A]):
    """
    `Empty` value for `Option`
    """

    def __init__(self, x: Any = None, e: Exception = None) -> None:
        self._reason = _Reason(x, e)

    def __repr__(self) -> str:
        return 'Empty()'

    def flat_map(self, f: Callable[[A], 'Option[B]']) -> 'Option[B]':
        return self

    def get(self):
        """Return the reason of the empty value.
        """
        return self._reason


empty = Empty()



# encoding=utf-8
# @Time    : 17-6-22
# @File    : __init__.py
# @Author  : jian<jian@mltalker.com>
from .operators import (
  DecodeImage, 
  ResizeS, 
  RandomFlipImage, 
  RandomDistort, 
  Rotation,
  KeepRatio, 
  ColorDistort,
  RandomErasingImage,
  RandomCropImageV1,
  RandomCropImageV2,
  MixupImage,
  CutmixImage,
  Meta,
  RandomScaledCrop,
  ResizeByLong,
  ResizeRangeScaling,
  ResizeStepScaling,
  AutoAugmentImage,
  Permute,
  UnSqueeze,
  ConvertRandomObjJointsAndOffset,
  ResizeByShort,
  RGB2BGR,
  FixedCrop,
  CorrectBoxes)

__all__ = [
  'DecodeImage', 
  'ResizeS',
  'RandomFlipImage', 
  'RandomDistort', 
  'Rotation', 
  'KeepRatio', 
  'ColorDistort',
  'RandomErasingImage',
  'RandomCropImageV1',
  'RandomCropImageV2',
  'Permute',
  'MixupImage',
  'CutmixImage',
  'RandomScaledCrop',
  'ResizeByLong',
  'ResizeRangeScaling',
  'ResizeStepScaling',
  'AutoAugmentImage',
  'UnSqueeze',
  'ConvertRandomObjJointsAndOffset',
  'ResizeByShort',
  'FixedCrop',
  'RGB2BGR',
  'CorrectBoxes',
  'Meta'
]



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:02
# @File    : dag.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from functools import wraps
from uuid import uuid4
from abc import ABCMeta

from antgo.pipeline.hparam import param_scope

def register_dag(f):
    @wraps(f)
    def wrapper(self, *arg, **kws):
        # Get the result DataCollections
        children = f(self, *arg, **kws)
        # Need the dc type while avoiding circular imports
        dc_type = type(children[0]) if isinstance(children, list) else type(children)
        # If the function is called from an existing dc
        index_info = param_scope()._index # pylint: disable=protected-access
        if index_info is None:
            input_info = None
            output_info = None
        elif isinstance(index_info, tuple):
            input_info = list(index_info[0]) if isinstance(index_info[0], tuple) else [index_info[0]]
            output_info = list(index_info[1]) if isinstance(index_info[1], tuple) else [index_info[1]]
        else:
            input_info = None
            output_info = list(index_info) if isinstance(index_info, tuple) else [index_info]
        if isinstance(self, dc_type):
            self.op = f.__name__
            self.call_args = {'*arg': arg, '*kws': kws}
            if arg != tuple():
                if hasattr(arg[0], '_op_config'):
                    self.op_config = arg[0].op_config
            # check if list of dc or just dc
            if isinstance(children, dc_type):
                self.child_ids = [children.id]
            else:
                self.child_ids = [x.id for x in children]
            info = {'op': self.op,
                    'op_name': self.op_name,
                    'is_stream': None,
                    'init_args': self.init_args,
                    'call_args': self.call_args,
                    'op_config': self.get_pipeline_config(),
                    'input_info': input_info,
                    'output_info': output_info,
                    'parent_ids': self.parent_ids,
                    'child_ids':  self.child_ids}
            # if has op_config, update op_config
            if self.op_config is not None:
                info['op_config'].update(self.op_config)
            self.get_control_plane().dag[self.id] = info
            return children
        # If not called from a dc, think static or class method.
        else:
            op = f.__name__
            # if the method is being called from a classmethod, avoiding passing in self
            if isinstance(self, (dc_type, ABCMeta)):
                pass_args = arg
            else:
                pass_args = (self,) + arg
            call_args = {'*arg': pass_args, '*kws': kws}
            if isinstance(children, dc_type):
                child_ids = [children.id]
            else:
                child_ids = [x.id for x in children]
            info = {'op': op,
                    'op_name': None,
                    'is_stream': None,
                    'init_args': None,
                    'call_args': call_args,
                    'op_config': None,
                    'input_info': input_info,
                    'output_info': output_info,
                    'parent_ids': [],
                    'child_ids':  child_ids}
            # If not called from a dc, it means that it is a start method
            # so it must be added to the childrens dags.
            for x in children if isinstance(children, list) else  [children]:
                x.get_control_plane().dag['start'] = info
            return children

    return wrapper


class DagMixin:
    #pylint: disable=import-outside-toplevel
    """
    Mixin for creating DAGs and their corresponding yamls from a DC
    """
    def __init__(self) -> None:
        super().__init__()
        # Unique id for current dc
        self.id = str(uuid4().hex[:8])
        with param_scope() as hp:
            parent = hp().data_collection.parent(None)
        if parent is None:
            self.parent_ids = ['start']
            self._control_plane = ControlPlane()
        else:
            self.parent_ids = [parent.id]
            self._control_plane = parent._control_plane
        self.op = None
        self.op_name = None
        self.init_args = None
        self.call_args = None
        self.op_config = None
        self.input_info = None
        self.output_info = None
        self.child_ids = []

    def register_dag(self, children):
        # check if list of dc or just dc
        if isinstance(children, type(self)):
            self.child_ids = [children.id]
        else:
            self.child_ids = [x.id for x in children]
        info = {'op': self.op,
                'op_name': self.op_name,
                'is_stream': self.is_stream,
                'init_args': self.init_args,
                'call_args': self.call_args,
                'op_config': self.op_config,
                'input_info': self.input_info,
                'output_info': self.output_info,
                'parent_ids': self.parent_ids,
                'child_ids':  self.child_ids}
        self._control_plane.dag[self.id] = info
        return children

    def notify_consumed(self, new_id):
        info = {'op': 'nop',
                'op_name': None,
                'init_args': None,
                'call_args': None,
                'op_config': None,
                'input_info': None,
                'output_info': None,
                'parent_ids': self.parent_ids,
                'child_ids': [new_id]}
        self._control_plane.dag[self.id] = info

    def compile_dag(self):
        info = {'op': 'nop',
                'op_name': None,
                'init_args': None,
                'call_args': None,
                'op_config': None,
                'input_info': None,
                'output_info': None,
                'parent_ids': self.parent_ids,
                'child_ids': ['end']}
        self._control_plane.dag[self.id] = info
        info = {'op': 'end',
                'op_name': None,
                'init_args': None,
                'call_args': None,
                'op_config': None,
                'input_info': None,
                'output_info': None,
                'parent_ids': [self.id],
                'child_ids': []}
        self._control_plane.dag['end'] = info
        # return self._control_plane.dag
        return self._clean_nops(self._control_plane.dag)

    def netx(self):
        import networkx as nx
        import matplotlib.pyplot as plt
        compiled_dag = self.compile_dag()
        new_dict = {}
        label_dict = {}
        for key,value in compiled_dag.items():
            # print(key, value)
            new_dict[key] = value['child_ids']
            label_dict[key] = value['op']
        label_dict['end'] = 'end'
        g = nx.DiGraph(new_dict)
        pos = nx.nx_pydot.graphviz_layout(g, 'dot')
        nx.draw_networkx(g, pos, labels=label_dict, with_labels = True)
        plt.show()

    def _clean_nops(self, dag):
        dag_copy = dag.copy()
        removals = []
        for key, val in dag.items():
            if val['op'] == 'nop':
                removals.append(key)
                for parent in val['parent_ids']:
                    dag_copy[parent]['child_ids'].remove(key)
                    dag_copy[parent]['child_ids'] = list(set(dag_copy[parent]['child_ids'] + val['child_ids']))
                for child in val['child_ids']:
                    dag_copy[child]['parent_ids'].remove(key)
                    dag_copy[child]['parent_ids'] = list(set(dag_copy[child]['parent_ids'] +val['parent_ids']))
        for x in removals:
            del dag_copy[x]
        return self._add_op_name_and_init_args(dag_copy)

    def _add_op_name_and_init_args(self, dag):
        for key, val in dag.items():
            if (val['op'] == 'map' or val['op'] == 'filter') or val['call_args'] is not None:
                if val['call_args']['*arg'] != () and hasattr(val['call_args']['*arg'][0], '_kws'):
                    dag[key]['init_args'] = val['call_args']['*arg'][0].init_args
                    if len(val['call_args']['*arg'][0].function.split('/')) > 1:
                        dag[key]['op_name'] = val['call_args']['*arg'][0].function
                    else:
                        dag[key]['op_name'] = 'antgo/' + val['call_args']['*arg'][0].function
                else:
                    dag[key]['op_name'] = 'dummy_input'
                    dag[key]['parent_ids'] = []
                    start = key
            else:
                dag[key]['op_name'] = 'end'
        dag['start'] = dag[start]
        del dag[start]
        return self._solve_child_ids(dag, start)

    def _solve_child_ids(self, dag, start):
        co_dag = dag.copy()
        for k, v in dag.items():
            if start in v['parent_ids']:
                co_dag[k]['parent_ids'].append('start')
                co_dag[k]['parent_ids'].remove(start)
        return self._solve_input_info(co_dag)

    def _solve_input_info(self, dag):
        copy_dag = dag.copy()
        for key, value in dag.items():
            if isinstance(value['input_info'], list):
                input_info = []
                for i in value['input_info']:
                    for j in value['parent_ids']:
                        if j == 'start':
                            input_info.append(tuple(['start', i]))
                        elif dag[j]['output_info'] is not None and (i in dag[j]['output_info']) :
                            input_info.append(tuple([j, i]))
                    copy_dag[key]['input_info'] = input_info
        return copy_dag

    def _clean_streams(self, dag):
        raise NotImplementedError

    def get_control_plane(self):
        return self._control_plane


class ControlPlane:
    def __init__(self) -> None:
        self._dag = {}

    @property
    def dag(self):
        return self._dag



from antgo.measures import *
from ..runner.builder import *
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval
from collections import defaultdict
import copy
import json
import numpy as np
import time


class COCOWarp(COCO):
    def __init__(self, gt_ann=None):
        """
        Constructor of Microsoft COCO helper class for reading and visualizing annotations.
        :param annotation_file (str): location of annotation file
        :param image_folder (str): location to the folder that hosts images.
        :return:
        """
        # load dataset
        self.dataset, self.anns, self.cats, self.imgs = dict(), dict(), dict(), dict()
        self.imgToAnns, self.catToImgs = defaultdict(list), defaultdict(list)
        if gt_ann is not None:
            print('loading annotations into memory...')
            tic = time.time()
            if isinstance(gt_ann, str):
                with open(gt_ann, 'r') as f:
                    dataset = json.load(f)
                assert type(dataset) == dict, 'annotation file format {} not supported'.format(type(dataset))
                print('Done (t={:0.2f}s)'.format(time.time() - tic))
            else:
                dataset = gt_ann
            self.dataset = dataset
            self.createIndex()    


@MEASURES.register_module()
class COCOCompatibleEval(object):
    def __init__(self, categories, without_background=True):
        self.categories = categories
        for c in self.categories:
            if 'supercategory' not in c:
                c['supercategory'] = 'default'
        
        self.without_background = without_background
    
    def keys(self):
        # 约束使用此评估方法，需要具体的关键字信息
        return {'pred': ['box', 'label'], 'gt': ['image_meta', 'bboxes', 'labels']}

    def __call__(self, preds, gts):
        # gts 格式 'info', 'licenses', 'images', 'annotations', 'categories’
        # 将GT 转换为COCO格式
        # image_metas, bboxes, labels
        images = []
        annotations = []
        bbox_id = 0
        for image_id, gt in enumerate(gts):
            image_file = gt['image_meta']['image_file'] if 'image_file' in gt['image_meta'] else ''
            bboxes = [[box[0], box[1], box[2]-box[0], box[3]-box[1]] for box in gt['bboxes'].tolist()]
            areas = [box[2]*box[3] for box in bboxes]
            category_ids = [l for l in gt['labels'].tolist()]

            for _, (bbox, area, category_id) in enumerate(zip(bboxes, areas, category_ids)):
                if self.without_background:
                    category_id += 1

                annotations.append({
                    "segmentation": [],
                    "iscrowd": 0,
                    "image_id": image_id+1,
                    "bbox":bbox,
                    "area": area,
                    "category_id": category_id,
                    "id": bbox_id+1,
                    'ignore': 0
                })
                bbox_id += 1

            images.append({
                'height': gt['image_meta']['image_shape'][0],
                'width': gt['image_meta']['image_shape'][1],
                'id': image_id+1,
                'file_name': image_file
            })

        gt_coco = COCOWarp({
            'images': images,
            'categories': self.categories,
            'annotations': annotations
        })

        # 将预测转换为COCO格式
        pred_annotations = []
        for image_id, pred in enumerate(preds):
            pred_bboxes = pred['box'][:,:4]
            pred_probs = pred['box'][:,4]
            pred_labels = pred['label']

            for _, (pred_bbox, pred_prob, pred_label) in enumerate(zip(pred_bboxes, pred_probs, pred_labels)):
                x1, y1, x2, y2 = pred_bbox
                score = (float)(pred_prob)
                label = (int)(pred_label)
                if self.without_background:
                    label += 1

                w, h = x2 - x1, y2 - y1
                pred_dict = {
                    "image_id": image_id+1,
                    "bbox": [ x1, y1, w, h ],
                    "category_id": (int)(label),
                    "score": score
                }
                pred_annotations.append(pred_dict)

        pred_coco = gt_coco.loadRes(pred_annotations)

        coco_eval = COCOeval(gt_coco, pred_coco, "bbox")
        coco_eval.evaluate()
        coco_eval.accumulate()
        coco_eval.summarize()
        tags = [
            'AP@[ IoU=0.50:0.95 | area= all | maxDets=100 ]',
            'AP@[ IoU=0.50 | area= all | maxDets=100 ]',
            'AP@[ IoU=0.75 | area= all | maxDets=100 ]',
            'AP@[ IoU=0.50:0.95 | area= small | maxDets=100 ]',
            'AP@[ IoU=0.50:0.95 | area=medium | maxDets=100 ]',
            'AP@[ IoU=0.50:0.95 | area= large | maxDets=100 ]',
            'AR@[ IoU=0.50:0.95 | area= all | maxDets= 1 ]',
            'AR@[ IoU=0.50:0.95 | area= all | maxDets= 10 ]',
            'AR@[ IoU=0.50:0.95 | area= all | maxDets=100 ]',
            'AR@[ IoU=0.50:0.95 | area= small | maxDets=100 ]',
            'AR@[ IoU=0.50:0.95 | area=medium | maxDets=100 ]',
            'AR@[ IoU=0.50:0.95 | area= large | maxDets=100 ]'
        ]

        tag_and_value = {}
        for tag, value in zip(tags, coco_eval.stats.tolist()):
            tag_and_value[tag] = value

        return tag_and_value


# cc = COCOBboxEval()
# gt_ann = '/root/paddlejob/workspace/env_run/portrait/COCO/annotations/instances_val2017.json'
# with open(gt_ann, 'r') as f:
#     dataset = json.load(f)

# category_map_id = {}
# categories = dataset['categories']
# for ci, c in enumerate(categories):
#     category_map_id[c['id']] = ci

# image_map_id = {}
# images = dataset['images']
# for image_i, image in enumerate(images):
#     image_map_id[image['id']] = image_i

# annotations = dataset['annotations']
# det_result = [[] for _ in range(len(images))]
# for ann in annotations:
#     image_i = image_map_id[ann['image_id']]
#     x,y,w,h = ann['bbox']
#     label = category_map_id[ann['category_id']]
#     det_result[image_i].append([x,y,x+w,y+h,0.5,label])

# cc(det_result,dataset)
# print('aa')
